"""
Stateful interface for accessing the current job (and store).

This module defines the ``CURRENT_JOB`` object which has two attributes:

- ``job``: Containing the current job.
- ``store``: Containing the current store. Only available if ``expose_store`` is set in
  the job config.
"""

from __future__ import annotations

import typing

from monty.design_patterns import singleton

if typing.TYPE_CHECKING:
    pass

    import jobflow

__all__ = ["CURRENT_JOB"]


@singleton
class State:
    """State of the current job and store."""

    job: jobflow.Job | None = None
    store: jobflow.JobStore | None = None

    def reset(self):
        """Reset the current state."""
        self.job = None
        self.store = None


CURRENT_JOB: State = State()



"""Utilities for enumerations."""

from enum import Enum

__all__ = ["ValueEnum"]


class ValueEnum(Enum):
    """Enum that serializes to string as the value and can be compared against a str."""

    def __str__(self):
        """Get a string representation of the enum."""
        return str(self.value)

    def __eq__(self, other):
        """Compare to another enum for equality."""
        if type(self) == type(other) and self.value == other.value:
            return True
        else:
            return str(self.value) == str(other)

    def __hash__(self):
        """Get a hash of the enum."""
        return hash(str(self))

    def as_dict(self):
        """Create a serializable representation of the enum."""
        return str(self.value)



"""Tools for running :obj:`Flow` and :obj:`Job` objects using the FireWorks package."""

from __future__ import annotations

import typing

from fireworks import FiretaskBase, Firework, FWAction, Workflow, explicit_serialize

if typing.TYPE_CHECKING:
    from typing import Sequence

    import jobflow

__all__ = ["flow_to_workflow", "job_to_firework", "JobFiretask"]


def flow_to_workflow(
    flow: jobflow.Flow | jobflow.Job | list[jobflow.Job],
    store: jobflow.JobStore | None = None,
    **kwargs,
) -> Workflow:
    """
    Convert a :obj:`Flow` or a :obj:`Job` to a FireWorks :obj:`Workflow` object.

    Each firework spec is updated with the contents of the
    :obj:`Job.config.manager_config` dictionary. Accordingly, a :obj:`.JobConfig` object
    can be used to configure FireWork options such as metadata and the fireworker.

    Parameters
    ----------
    flow
        A flow or job.
    store
        A job store. Alternatively, if set to None, :obj:`JobflowSettings.JOB_STORE`
        will be used. Note, this could be different on the computer that submits the
        workflow and the computer which runs the workflow. The value of ``JOB_STORE`` on
        the computer that runs the workflow will be used.
    **kwargs
        Keyword arguments passed to Workflow init method.

    Returns
    -------
    Workflow
        The job or flow as a workflow.
    """
    from fireworks.core.firework import Firework, Workflow

    from jobflow.core.flow import get_flow

    parent_mapping: dict[str, Firework] = {}
    fireworks = []

    flow = get_flow(flow)

    for job, parents in flow.iterflow():
        fw = job_to_firework(job, store, parents=parents, parent_mapping=parent_mapping)
        fireworks.append(fw)

    return Workflow(fireworks, name=flow.name, **kwargs)


def job_to_firework(
    job: jobflow.Job,
    store: jobflow.JobStore | None = None,
    parents: Sequence[str] | None = None,
    parent_mapping: dict[str, Firework] | None = None,
    **kwargs,
) -> Firework:
    """
    Convert a :obj:`Job` to a :obj:`.Firework`.

    The firework spec is updated with the contents of the
    :obj:`Job.config.manager_config` dictionary. Accordingly, a :obj:`.JobConfig` object
    can be used to configure FireWork options such as metadata and the fireworker.

    Parameters
    ----------
    job
        A job.
    store
        A job store. Alternatively, if set to None, :obj:`JobflowSettings.JOB_STORE`
        will be used. Note, this could be different on the computer that submits the
        workflow and the computer which runs the workflow. The value of ``JOB_STORE`` on
        the computer that runs the workflow will be used.
    parents
        The parent uuids of the job.
    parent_mapping
        A dictionary mapping job uuids to Firework objects, as ``{uuid: Firework}``.
    **kwargs
        Keyword arguments passed to the Firework constructor.

    Returns
    -------
    Firework
        A firework that will run the job.
    """
    from fireworks.core.firework import Firework

    from jobflow.core.reference import OnMissing

    if (parents is None) is not (parent_mapping is None):
        raise ValueError("Both or neither of parents and parent_mapping must be set.")

    task = JobFiretask(job=job, store=store)

    job_parents = None
    if parents is not None and parent_mapping is not None:
        job_parents = (
            [parent_mapping[parent] for parent in parents] if parents else None
        )

    spec = {"_add_launchpad_and_fw_id": True}  # this allows the job to know the fw_id
    if job.config.on_missing_references != OnMissing.ERROR:
        spec["_allow_fizzled_parents"] = True
    spec.update(job.config.manager_config)
    spec.update(job.metadata)  # add metadata to spec

    fw = Firework([task], spec=spec, name=job.name, parents=job_parents, **kwargs)

    if parent_mapping is not None:
        parent_mapping[job.uuid] = fw

    return fw


@explicit_serialize
class JobFiretask(FiretaskBase):
    """
    A firetask that will run any job.

    Other Parameters
    ----------------
    job : Dict
        A serialized job.
    store : JobStore
        A job store. Alternatively, if set to None, :obj:`JobflowSettings.JOB_STORE`
        will be used. Note, this could be different on the computer that submits the
        workflow and the computer which runs the workflow. The value of ``JOB_STORE`` on
        the computer that runs the workflow will be used.
    """

    required_params = ["job", "store"]

    def run_task(self, fw_spec):
        """Run the job and handle any dynamic firework submissions."""
        from jobflow import SETTINGS, initialize_logger
        from jobflow.core.job import Job

        job: Job = self.get("job")
        store = self.get("store")

        if store is None:
            store = SETTINGS.JOB_STORE
        store.connect()

        if hasattr(self, "fw_id"):
            job.metadata.update({"fw_id": self.fw_id})

        initialize_logger()
        response = job.run(store=store)

        detours = None
        additions = None
        if response.replace is not None:
            # create a workflow from the new additions; be sure to use original store
            detours = [flow_to_workflow(response.replace, self.get("store"))]

        if response.addition is not None:
            additions = [flow_to_workflow(response.addition, self.get("store"))]

        if response.detour is not None:
            detour_wf = flow_to_workflow(response.detour, self.get("store"))
            if detours is not None:
                detours.append(detour_wf)
            else:
                detours = [detour_wf]

        fwa = FWAction(
            stored_data=response.stored_data,
            detours=detours,
            additions=additions,
            defuse_workflow=response.stop_jobflow,
            defuse_children=response.stop_children,
        )
        return fwa



"""Jobflow is a package for writing dynamic and connected workflows."""

from jobflow._version import __version__
from jobflow.core.flow import Flow, JobOrder
from jobflow.core.job import Job, JobConfig, Response, job
from jobflow.core.maker import Maker
from jobflow.core.reference import OnMissing, OutputReference
from jobflow.core.state import CURRENT_JOB
from jobflow.core.store import JobStore
from jobflow.managers.local import run_locally
from jobflow.settings import JobflowSettings
from jobflow.utils.log import initialize_logger

SETTINGS = JobflowSettings()



"""Tools for logging."""

import logging

__all__ = ["initialize_logger"]


def initialize_logger(level: int = logging.INFO) -> logging.Logger:
    """Initialize the default logger.

    Parameters
    ----------
    level
        The log level.

    Returns
    -------
    Logger
        A logging instance with customized formatter and handlers.
    """
    import sys

    log = logging.getLogger("jobflow")
    log.setLevel(level)
    log.handlers = []  # reset logging handlers if they already exist

    fmt = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(fmt)
    log.addHandler(screen_handler)
    return log



"""Tools for running jobflow locally."""

from __future__ import annotations

import logging
import typing

if typing.TYPE_CHECKING:
    pass

    import jobflow

__all__ = ["run_locally"]

logger = logging.getLogger(__name__)


def run_locally(
    flow: jobflow.Flow | jobflow.Job | list[jobflow.Job],
    log: bool = True,
    store: jobflow.JobStore | None = None,
    create_folders: bool = False,
    ensure_success: bool = False,
) -> dict[str, dict[int, jobflow.Response]]:
    """
    Run a :obj:`Job` or :obj:`Flow` locally.

    Parameters
    ----------
    flow
        A job or flow.
    log
        Whether to print log messages.
    store
        A job store. If a job store is not specified then
        :obj:`JobflowSettings.JOB_STORE` will be used. By default this is a maggma
        ``MemoryStore`` but can be customised by setting the jobflow configuration file.
    create_folders
        Whether to run each job in a new folder.
    ensure_success
        Raise an error if the flow was not executed successfully.

    Returns
    -------
    Dict[str, Dict[int, Response]]
        The responses of the jobs, as a dict of ``{uuid: {index: response}}``.
    """
    from collections import defaultdict
    from datetime import datetime
    from pathlib import Path
    from random import randint

    from monty.os import cd

    from jobflow import SETTINGS, initialize_logger
    from jobflow.core.flow import get_flow
    from jobflow.core.reference import OnMissing

    if store is None:
        store = SETTINGS.JOB_STORE

    store.connect()

    if log:
        initialize_logger()

    flow = get_flow(flow)

    stopped_parents: set[str] = set()
    errored: set[str] = set()
    responses: dict[str, dict[int, jobflow.Response]] = defaultdict(dict)
    stop_jobflow = False

    root_dir = Path.cwd()

    def _run_job(job: jobflow.Job, parents):
        nonlocal stop_jobflow

        if stop_jobflow:
            return False

        if len(set(parents).intersection(stopped_parents)) > 0:
            # stop children has been called for one of the jobs' parents
            logger.info(
                f"{job.name} is a child of a job with stop_children=True, skipping..."
            )
            stopped_parents.add(job.uuid)
            return

        if (
            len(set(parents).intersection(errored)) > 0
            and job.config.on_missing_references == OnMissing.ERROR
        ):
            errored.add(job.uuid)
            return

        try:
            response = job.run(store=store)
        except Exception:
            import traceback

            logger.info(f"{job.name} failed with exception:\n{traceback.format_exc()}")
            errored.add(job.uuid)
            return

        responses[job.uuid][job.index] = response

        if response.stored_data is not None:
            logger.warning("Response.stored_data is not supported with local manager.")

        if response.stop_children:
            stopped_parents.add(job.uuid)

        if response.stop_jobflow:
            stop_jobflow = True
            return False

        if response.replace is not None:
            # first run any restarts
            _run(response.replace)

        if response.detour is not None:
            # next any detours
            _run(response.detour)

        if response.addition is not None:
            # finally any additions
            _run(response.addition)

        return response

    def _get_job_dir():
        if create_folders:
            time_now = datetime.utcnow().strftime(SETTINGS.DIRECTORY_FORMAT)
            job_dir = root_dir / f"job_{time_now}-{randint(10000, 99999)}"
            job_dir.mkdir()
            return job_dir
        else:
            return root_dir

    def _run(root_flow):
        job: jobflow.Job
        for job, parents in root_flow.iterflow():
            job_dir = _get_job_dir()
            with cd(job_dir):
                response = _run_job(job, parents)
            if response is False:
                return False

        return response is not None

    logger.info("Started executing jobs locally")
    finished_successfully = _run(flow)
    logger.info("Finished executing jobs locally")

    if ensure_success and not finished_successfully:
        raise RuntimeError("Flow did not finish running successfully")

    return dict(responses)



"""Define base Maker class for creating jobs and flows."""

from __future__ import annotations

import typing
from dataclasses import dataclass

from monty.json import MontyDecoder, MSONable

if typing.TYPE_CHECKING:
    from typing import Any, Callable

    import jobflow

__all__ = ["Maker"]


@dataclass
class Maker(MSONable):
    """
    Base maker (factory) class for constructing :obj:`Job` and :obj:`Flow` objects.

    Note, this class is only an abstract implementation. To create a functioning Maker,
    one should  subclass :obj:`Maker`, define the ``make`` method and add a name field
    with a default value. See examples below.

    See Also
    --------
    .job, .Job, .Flow

    Examples
    --------
    Below we define a simple maker that generates a job to add two numbers. Next, we
    make a job using the maker. When creating a new maker it is essential to: 1)
    Add a name field with a default value (this controls the name of jobs produced
    by the maker) and 2) override the ``make`` method with a concrete implementation.

    >>> from dataclasses import dataclass
    >>> from jobflow import Maker, job
    >>> @dataclass
    ... class AddMaker(Maker):
    ...     name: str = "add"
    ...
    ...     @job
    ...     def make(self, a, b):
    ...         return a + b
    >>> maker = AddMaker()
    >>> add_job = maker.make(1, 2)
    >>> add_job.name
    "add"

    There are two key features of :obj:`Maker` objects that make them extremely
    powerful. The first is the ability to have class instance variables that
    are used inside the make function. For example, running the job below will produce
    the output ``6.5``.

    >>> @dataclass
    ... class AddMaker(Maker):
    ...     name: str = "add"
    ...     c: float = 1.5
    ...
    ...     @job
    ...     def make(self, a, b):
    ...         return a + b + self.c
    >>> maker = AddMaker(c=3.5)
    >>> add_job = maker.make(1, 2)

    .. Note::
        Jobs created by a Maker will inherit a copy of the Maker object. This means that
        if 1) a job is created, 2) the maker variables are changed, and 3) a new job is
        created, only the second job will reflect the updated maker variables.

    The second useful feature is the ability for delayed creation of jobs. For example,
    consider a job that creates another job during its execution. If the new job
    has lots of optional arguments, this will require passing lots of arguments to the
    first job. However, using a Maker, the kwargs of the job can be specified as maker
    kwargs and specified when the first job is created, making it much cleaner to
    customise the newly created job. For example, consider this job that first squares
    two numbers, then creates a new job to add them.

    >>> from jobflow import Response
    >>> @dataclass
    ... class SquareThenAddMaker(Maker):
    ...     name: str = "square then add"
    ...     add_maker: Maker = AddMaker()
    ...
    ...     @job
    ...     def make(self, a, b):
    ...         a_squared = a ** 2
    ...         b_squared = b ** 2
    ...         add_job = self.add_maker.make(a_squared, b_squared)
    ...         return Response(detour=add_job)
    >>> maker = SquareThenAddMaker()
    >>> square_add_job = maker.make(3, 4)

    The add job can be customised by specifying a custom ``AddMaker`` instance when
    the ``SquareThenAddMaker`` is initialized.

    >>> custom_add_maker = AddMaker(c=1000)
    >>> maker = SquareThenAddMaker(add_maker=custom_add_maker)

    This provides a very natural way to contain and control the kwargs for dynamically
    created jobs.

    Note that makers can also be used to create :obj:`Flow` objects. In this case,
    the only difference is that the ``make()`` method is not decorated using the ``job``
    decorator and the function should return a flow. For example:

    >>> from jobflow import Flow
    >>> @dataclass
    ... class DoubleAddMaker(Maker):
    ...     name: str = "double add"
    ...     add_maker: Maker = AddMaker()
    ...
    ...     def make(self, a, b):
    ...         add_first = self.add_maker.make(a, b)
    ...         add_second = self.add_maker.make(add_first.output, b)
    ...         return Flow([add_first, add_second], name=self.name)
    >>> maker = DoubleAddMaker()
    >>> double_add_job = maker.make(1, 2)
    """

    def make(self, *args, **kwargs) -> jobflow.Flow | jobflow.Job:
        """Make a job or a flow - must be overridden with a concrete implementation."""
        raise NotImplementedError

    @property
    def name(self):
        """Name of the Maker - must be overridden with a dataclass field."""
        raise NotImplementedError

    def update_kwargs(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        class_filter: type[Maker] | None = None,
        nested: bool = True,
        dict_mod: bool = False,
    ):
        """
        Update the keyword arguments of the :obj:`.Maker`.

        .. Note::
            Note, the maker is not updated in place. Instead a new maker object is
            returned.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the Maker name. Only Makers with a matching name will be
            updated. Includes partial matches, e.g. "ad" will match a Maker with the
            name "adder".
        class_filter
            A filter for the maker class. Only Makers with a matching class will be
            updated. Note the class filter will match any subclasses.
        nested
            Whether to apply the updates to Maker objects that are themselves kwargs
            of a Maker object. See examples for more details.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.

        Examples
        --------
        Consider the following Maker:

        >>> from dataclasses import dataclass
        >>> from jobflow import job, Maker
        >>> @dataclass
        ... class AddMaker(Maker):
        ...     name: str = "add"
        ...     number: float = 10
        ...
        ...     @job
        ...     def make(self, a):
        ...         return a + self.number
        >>> maker = AddMaker()

        The ``number`` argument could be updated using:

        >>> maker.update_kwargs({"number": 10})

        By default, the updates are applied to nested Makers. These are Makers
        which are present in the kwargs of another Maker. For example, consider this
        maker that first squares a number, then creates a new job to add it.

        >>> from jobflow import Response
        >>> @dataclass
        ... class SquareThenAddMaker(Maker):
        ...     name: str = "square then add"
        ...     add_maker: Maker = AddMaker()
        ...
        ...     @job
        ...     def make(self, a, b):
        ...         a_squared = a ** 2
        ...         add_job = self.add_maker.make(a_squared)
        ...         return Response(detour=add_job)
        >>> maker = SquareThenAddMaker()

        The following update will apply to the nested ``AddMaker`` in the kwargs of the
        ``SquareThenAddMaker``:

        >>> maker = maker.update_kwargs({"number": 10}, class_filter=AddMaker)

        However, if ``nested=False``, then the update will not be applied to the nested
        Maker:

        >>> maker = maker.update_kwargs(
        ...     {"number": 10}, class_filter=AddMaker, nested=False
        ... )
        """
        from jobflow.utils.dict_mods import apply_mod

        def _update_kwargs_func(maker: Maker):
            # Update the kwargs of the maker
            d = maker.as_dict()
            if dict_mod:
                apply_mod(update, d)
            else:
                d.update(update)
            return maker.from_dict(d)

        return recursive_call(
            self,
            func=_update_kwargs_func,
            name_filter=name_filter,
            class_filter=class_filter,
            nested=nested,
        )


def recursive_call(
    obj: Maker,
    func: Callable[[Maker], Maker],
    name_filter: str | None = None,
    class_filter: type[Maker] | None = None,
    nested: bool = True,
):
    """Recursively call a function on all Maker objects in the object.

    Parameters
    ----------
    obj
        The Maker object to call the function on.
    func
        The function to call a Maker object, if it matches the filters.
        This function must return a new Maker object.
    name_filter
        A filter for the Maker name. Only Makers with a matching name will be updated.
        Includes partial matches, e.g. "ad" will match a Maker with the name "adder".
    class_filter
        A filter for the maker class. Only Makers with a matching class will be
        updated. Note the class filter will match any subclasses.
    nested
        Whether to apply the updates to Maker objects that are themselves kwargs
        of a Maker object. See examples for more details.
    """
    from pydash import get, set_

    from jobflow.utils.find import find_key

    if isinstance(class_filter, Maker):
        # Maker instance supplied rather than a Maker class
        class_filter = class_filter.__class__

    def _filter(nested_obj: Maker):
        # Filter the Maker object
        if not isinstance(nested_obj, Maker):
            return False
        if name_filter is not None and name_filter not in nested_obj.name:
            return False
        if class_filter is not None and not isinstance(nested_obj, class_filter):
            return False
        return True

    d = obj.as_dict()

    # find and update makers in Maker kwargs. Process is:
    # 1. Look for any monty classes in serialized maker kwargs
    # 2. Regenerate the classes and check if they are a Maker
    # 3. Call the functions on the deepest classes first
    # 4. Call the function or update on the deepest classes and move up the tree
    # 5. Finally, replace the call/update the object itself

    # find all classes in the serialized maker kwargs
    # will only look at the top level if nested=False
    locations = find_key(d, "@class", nested=True) if nested else [[]]

    for location in sorted(
        locations, key=len, reverse=True
    ):  # should deserialize in order
        if len(location) == 0:
            continue
        nested_class = MontyDecoder().process_decoded(get(d, list(location)))
        if _filter(nested_class):
            # either update or call the function on the nested Maker
            modified_class = func(nested_class)
            if not isinstance(modified_class, Maker):
                raise ValueError(
                    "Function must return a Maker object. "
                    f"Got {type(modified_class)} instead."
                )
            # update the serialized maker with the new kwarg
            set_(d, list(location), modified_class.as_dict())

    # the top level must be processed separately since its constructor
    # might not be discoverable by MontyDecoder (kinda hacky)
    new_obj = obj.from_dict(d)
    if _filter(obj):
        new_obj = func(new_obj)
    return new_obj



"""Define classes for handling job outputs."""

from __future__ import annotations

import contextlib
import typing
from typing import Any, Sequence

from monty.json import MontyDecoder, MontyEncoder, MSONable, jsanitize
from pydantic import BaseModel
from pydantic.utils import lenient_issubclass

from jobflow.utils.enum import ValueEnum

if typing.TYPE_CHECKING:

    import jobflow

__all__ = [
    "OnMissing",
    "OutputReference",
    "resolve_references",
    "find_and_resolve_references",
    "find_and_get_references",
]


class OnMissing(ValueEnum):
    """
    What to do when a reference cannot be resolved.

    - ``ERROR``: Throw an error.
    - ``NONE``: Replace the missing reference with ``None``.
    - ``PASS``: Pass the unresolved :obj:`OutputReference` object on.
    """

    ERROR = "error"
    NONE = "none"
    PASS = "pass"


class OutputReference(MSONable):
    """
    A reference to the output of a :obj:`Job`.

    The output may or may not exist yet. Accordingly, :obj:`OutputReference` objects
    are essentially pointers to future job outputs. Output references can be used as
    input to a :obj:`Job` or :obj:`Flow` and will be "resolved". This means, the value
    of the reference will be looked up in the job store and the value passed on to the
    job.

    Output references are also how jobflow is able to determine the dependencies between
    jobs, allowing for construction of the job graph and to determine the job execution
    order.

    Upon attribute access or indexing, an output reference will return another
    output reference with the attribute accesses stored. When resolving the reference,
    the attributes/indexing will be also be performed before the value is passed on to
    the job. See the examples for more details.

    Parameters
    ----------
    uuid
        The job uuid to which the output belongs.
    attributes :
        A tuple of attributes or indexes that have been performed on the
        output. Attributes are specified by a tuple of ``("a", attr)``, whereas
        indexes are specified as a tuple of ``("i", index)``.
    output_schema
        A pydantic model that defines the schema of the output and that will be used to
        validate any attribute accesses or indexes. Note, schemas can only be used to
        validate the first attribute/index access.

    Examples
    --------
    A reference can be constructed using just the job uuid.

    >>> from jobflow import OutputReference
    >>> ref = OutputReference("1234")

    Attribute accesses return new references.

    >>> ref.my_attribute
    OutputReference(1234, .my_attribute)

    Attribute accesses and indexing can be chained.

    >>> ref["key"][0].value
    OutputReference(1234, ['key'], [0], .value)
    """

    __slots__ = ("uuid", "attributes", "output_schema")

    def __init__(
        self,
        uuid: str,
        attributes: tuple[tuple[str, Any], ...] = (),
        output_schema: type[BaseModel] | None = None,
    ):
        super().__init__()
        self.uuid = uuid
        self.attributes = attributes
        self.output_schema = output_schema

        for attr_type, attr in attributes:
            if attr_type not in ("a", "i"):
                raise ValueError(
                    f"Unrecognised attribute type '{attr_type}' for attribute '{attr}'"
                )

    def resolve(
        self,
        store: jobflow.JobStore | None,
        cache: dict[str, Any] | None = None,
        on_missing: OnMissing = OnMissing.ERROR,
    ) -> Any:
        """
        Resolve the reference.

        This function will query the job store for the reference value and perform any
        attribute accesses/indexing.

        .. Note::
            When resolving multiple references simultaneously it is more efficient to
            use :obj:`.resolve_references` as it will minimize the number of database
            requests.

        Parameters
        ----------
        store
            A job store.
        cache
            A dictionary cache to use for local caching of reference values.
        on_missing
            What to do if the output reference is missing in the database and cache.
            See :obj:`OnMissing` for the available options.

        Raises
        ------
        ValueError
            If the reference cannot be found and ``on_missing`` is set to
            ``OnMissing.ERROR`` (default).

        Returns
        -------
        Any
            The resolved reference if it can be found. If the reference cannot be found,
            the returned value will depend on the value of ``on_missing``.
        """
        if cache is None:
            cache = {}

        if self.uuid not in cache:
            cache[self.uuid] = {}

        # get the latest index for the output
        result = store.query_one({"uuid": self.uuid}, ["index"], sort={"index": -1})
        index = None if result is None else result["index"]

        if index is not None and index not in cache[self.uuid]:
            with contextlib.suppress(ValueError):
                cache[self.uuid][index] = store.get_output(
                    self.uuid, which="last", load=True, on_missing=on_missing
                )

        if on_missing == OnMissing.ERROR and index not in cache[self.uuid]:
            istr = f" ({index})" if index is not None else ""
            raise ValueError(
                f"Could not resolve reference - {self.uuid}{istr} not in store or cache"
            )
        elif on_missing == OnMissing.NONE and index not in cache[self.uuid]:
            return None
        elif on_missing == OnMissing.PASS and index not in cache[self.uuid]:
            return self

        data = cache[self.uuid][index]

        # decode objects before attribute access
        data = MontyDecoder().process_decoded(data)

        # re-cache data in case other references need it
        cache[self.uuid][index] = data

        for attr_type, attr in self.attributes:
            # i means index else use attribute access
            data = data[attr] if attr_type == "i" else getattr(data, attr)

        return data

    def set_uuid(self, uuid: str, inplace=True) -> OutputReference:
        """
        Set the UUID of the reference.

        Parameters
        ----------
        uuid
            A new UUID.
        inplace
            Whether to update the current reference object or return a completely new
            object.

        Returns
        -------
        OutputReference
            An output reference with the specified uuid.
        """
        if inplace:
            self.uuid = uuid
            return self
        else:
            from copy import deepcopy

            new_reference = deepcopy(self)
            new_reference.uuid = uuid
            return new_reference

    def __getitem__(self, item) -> OutputReference:
        """Index the reference."""
        subschema = None
        if self.output_schema is not None:
            _, subschema = validate_schema_access(self.output_schema, item)

        return OutputReference(
            self.uuid, (*self.attributes, ("i", item)), output_schema=subschema
        )

    def __getattr__(self, item) -> OutputReference:
        """Attribute access of the reference."""
        if item in {"kwargs", "args", "schema"} or (
            isinstance(item, str) and item.startswith("__")
        ):
            # This is necessary to trick monty/pydantic.
            raise AttributeError

        subschema = None
        if self.output_schema is not None:
            _, subschema = validate_schema_access(self.output_schema, item)

        return OutputReference(
            self.uuid, (*self.attributes, ("a", item)), output_schema=subschema
        )

    def __setattr__(self, attr, val):
        """Set attribute."""
        # Setting unknown attributes is not allowed
        if attr in self.__slots__:
            object.__setattr__(self, attr, val)
        else:
            raise TypeError("OutputReference objects are immutable")

    def __iter__(self):
        """Make sure OutputReference is not iterable."""
        raise TypeError("OutputReference objects are not iterable")

    def __setitem__(self, index, val):
        """Set item."""
        # Setting items via indexing is not allowed.
        raise TypeError("OutputReference objects are immutable")

    def __repr__(self) -> str:
        """Get a string representation of the reference and attributes."""
        if len(self.attributes) > 0:
            attribute_str = ", " + ", ".join(self.attributes_formatted)
        else:
            attribute_str = ""

        return f"OutputReference({str(self.uuid)}{attribute_str})"

    def __hash__(self) -> int:
        """Return a hash of the reference."""
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        """Test for equality against another reference."""
        if isinstance(other, OutputReference):
            return (
                self.uuid == other.uuid
                and len(self.attributes) == len(other.attributes)
                and all(
                    [
                        a[0] == b[0] and a[1] == b[1]
                        for a, b in zip(self.attributes, other.attributes)
                    ]
                )
            )
        return False

    @property
    def attributes_formatted(self):
        """Get a formatted description of the attributes."""
        return [
            f".{x[1]}" if x[0] == "a" else f"[{repr(x[1])}]" for x in self.attributes
        ]

    def as_dict(self):
        """Serialize the reference as a dict."""
        schema = self.output_schema
        schema_dict = MontyEncoder().default(schema) if schema is not None else None
        data = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "@version": None,
            "uuid": self.uuid,
            "attributes": self.attributes,
            "output_schema": schema_dict,
        }
        return data


def resolve_references(
    references: Sequence[OutputReference],
    store: jobflow.JobStore,
    cache: dict[str, Any] | None = None,
    on_missing: OnMissing = OnMissing.ERROR,
) -> dict[OutputReference, Any]:
    """
    Resolve multiple output references.

    Uses caching to minimize number of database queries.

    Parameters
    ----------
    references :
        A list or tuple of output references.
    store
        A job store.
    cache
        A dictionary cache to use for local caching of reference values.
    on_missing
        What to do if the output reference is missing in the database and cache.
        See :obj:`OnMissing` for the available options.

    Returns
    -------
    dict[OutputReference, Any]
        The output values as a dictionary mapping of ``{reference: output}``.
    """
    from itertools import groupby

    resolved_references = {}
    if cache is None:
        cache = {}

    for uuid, ref_group in groupby(references, key=lambda x: x.uuid):
        # get latest index
        result = store.query_one({"uuid": uuid}, ["index"], sort={"index": -1})
        index = None if result is None else result["index"]

        if uuid not in cache:
            cache[uuid] = {}

        if index is not None and index not in cache[uuid]:
            cache[uuid][index] = store.get_output(
                uuid, load=True, on_missing=on_missing
            )

        for ref in ref_group:
            resolved_references[ref] = ref.resolve(
                store, cache=cache, on_missing=on_missing
            )

    return resolved_references


def find_and_get_references(arg: Any) -> tuple[OutputReference, ...]:
    """
    Find and extract output references.

    This function works on nested inputs. For example, lists or dictionaries
    (or combinations of list and dictionaries) that contain output references.

    Parameters
    ----------
    arg
        The argument to search for references.

    Returns
    -------
    tuple[OutputReference]
        The output references as a tuple.
    """
    from pydash import get

    from jobflow.utils.find import find_key_value

    if isinstance(arg, OutputReference):
        # if the argument is a reference then stop there
        return (arg,)

    elif isinstance(arg, (float, int, str, bool)):
        # argument is a primitive, we won't find a reference here
        return ()

    arg = jsanitize(arg, strict=True, enum_values=True, allow_bson=True)

    # recursively find any reference classes
    locations = find_key_value(arg, "@class", "OutputReference")

    # deserialize references and return
    return tuple(OutputReference.from_dict(get(arg, loc)) for loc in locations)


def find_and_resolve_references(
    arg: Any,
    store: jobflow.JobStore,
    cache: dict[str, Any] | None = None,
    on_missing: OnMissing = OnMissing.ERROR,
) -> Any:
    """
    Return the input but with all output references replaced with their resolved values.

    This function works on nested inputs. For example, lists or dictionaries
    (or combinations of list and dictionaries) that contain output references.

    Parameters
    ----------
    arg
        The input argument containing output references.
    store
        A job store.
    cache
        A dictionary cache to use for local caching of reference values.
    on_missing
        What to do if the output reference is missing in the database and cache.
        See :obj:`OnMissing` for the available options.

    Returns
    -------
    Any
        The input argument but with all output references replaced with their resolved
        values. If a reference cannot be found, its replacement value will depend on the
        value of ``on_missing``.
    """
    from pydash import get, set_

    from jobflow.utils.find import find_key_value

    if isinstance(arg, dict) and arg.get("@class") == "OutputReference":
        # if arg is a deserialized reference, serialize it
        arg = OutputReference.from_dict(arg)

    if isinstance(arg, OutputReference):
        # if the argument is a reference then stop there
        return arg.resolve(store, cache=cache, on_missing=on_missing)

    elif isinstance(arg, (float, int, str, bool)):
        # argument is a primitive, we won't find a reference here
        return arg

    # serialize the argument to a dictionary
    encoded_arg = jsanitize(arg, strict=True, enum_values=True, allow_bson=True)

    # recursively find any reference classes
    locations = find_key_value(encoded_arg, "@class", "OutputReference")

    if len(locations) == 0:
        return arg

    # resolve the references
    references = [
        OutputReference.from_dict(get(encoded_arg, list(loc))) for loc in locations
    ]
    resolved_references = resolve_references(
        references, store, cache=cache, on_missing=on_missing
    )

    # replace the references in the arg dict
    for location, reference in zip(locations, references):
        # skip references that have not been resolved, e.g., on missing is PASS
        if reference == resolved_references[reference]:
            continue

        resolved_reference = resolved_references[reference]
        set_(encoded_arg, list(location), resolved_reference)

    # deserialize dict array
    return MontyDecoder().process_decoded(encoded_arg)


def validate_schema_access(
    schema: type[BaseModel], item: str
) -> tuple[bool, BaseModel | None]:
    """
    Validate that an attribute or index access is supported by a model.

    If the item is associated to a nested model the class is returned.

    Parameters
    ----------
    schema
        A pydantic model to use as the schema.
    item
        An attribute or key to access.

    Raises
    ------
    AttributeError
        If the item is not a valid attribute of the schema.

    Returns
    -------
    tuple[bool, BaseModel]
        the bool is ``True`` if the schema access was valid.
        The BaseModel class associated with the item, if any.
    """
    schema_dict = schema.schema()
    if item not in schema_dict["properties"]:
        raise AttributeError(f"{schema.__name__} does not have attribute '{item}'.")

    subschema = None
    item_type = schema.__fields__[item].outer_type_
    if lenient_issubclass(item_type, BaseModel):
        subschema = item_type

    return True, subschema



"""Settings for jobflow."""

from collections import defaultdict
from pathlib import Path

from maggma.stores import MemoryStore
from pydantic import BaseSettings, Field, root_validator

from jobflow import JobStore

DEFAULT_CONFIG_FILE_PATH = Path("~/.jobflow.yaml").expanduser().as_posix()

__all__ = ["JobflowSettings"]


def _default_additional_store():
    """Create a default MemoryStore and connect it.

    This is a private function used for the additional_stores in
    the default JOB_STORE.
    """
    mem_store = MemoryStore()
    mem_store.connect()
    return mem_store


class JobflowSettings(BaseSettings):
    """
    Settings for jobflow.

    The default way to modify these is to modify ~/.jobflow.yaml. Alternatively,
    the environment variable ``JOBFLOW_CONFIG_FILE`` can be set to point to a yaml file
    with jobflow settings.

    Lastly, the variables can be modified directly though environment variables by
    using the "JOBFLOW" prefix. E..g., ``JOBFLOW_JOB_STORE=path/to/jobstore.file``.

    **Allowed JOB_STORE formats**

    If the store is not supplied, a ``MemoryStore`` will be used. Can be specified in
    multiple formats.

    The simplest format is the yaml dumped version of the store, generated using:

    >>> import yaml
    >>> yaml.dump(store.as_dict())

    Alternatively, the store can be specified as the keys docs_store, additional_stores
    and any other keyword arguments supported by the :obj:`JobStore` constructor. The
    docs_store and additional stores are specified by the ``type`` key which must match
    a Maggma ``Store`` subclass, and the remaining keys are passed to the store
    constructor. For example, the following file would  create a :obj:`JobStore` with a
    ``MongoStore`` for docs and a ``GridFSStore`` or ``S3Store`` as an additional store
    for data.

    GridFSStore example:

    .. code-block:: yaml

        docs_store:
          type: MongoStore
          database: jobflow_unittest
          collection_name: outputs
          host: localhost
          port: 27017
        additional_stores:
          data:
            type: GridFSStore
            database: jobflow_unittest
            collection_name: outputs_blobs
            host: localhost
            port: 27017

    S3Store example (Note: the ``key`` field must be set to ``blob_uuid``):

    .. code-block:: yaml

        docs_store:
          type: MongoStore
          database: jobflow_unittest
          collection_name: outputs
          host: localhost
          port: 27017
        additional_stores:
          data:
            type: S3Store
            bucket: output_blobs
            key: blob_uuid
            index:
              type: MongoStore
              database: jobflow_unittest
              collection_name: output_blobs_index
              host: localhost
              port: 27017
              key: blob_uuid


    Lastly, the store can be specified as a file name that points to a file containing
    the credentials in any format supported by :obj:`.JobStore.from_file`.
    """

    CONFIG_FILE: str = Field(
        DEFAULT_CONFIG_FILE_PATH, description="File to load alternative defaults from."
    )

    # general settings
    JOB_STORE: JobStore = Field(
        default_factory=lambda: JobStore(
            MemoryStore(),
            additional_stores=defaultdict(lambda: _default_additional_store()),
        ),
        description="Default JobStore to use when running locally or using FireWorks. "
        "See the :obj:`JobflowSettings` docstring for more details on the "
        "accepted formats.",
    )
    DIRECTORY_FORMAT: str = Field(
        "%Y-%m-%d-%H-%M-%S-%f",
        description="Date stamp format used to create directories",
    )

    class Config:
        """Pydantic config settings."""

        env_prefix = "jobflow_"

    @root_validator(pre=True)
    def load_default_settings(cls, values):
        """
        Load settings from file or environment variables.

        Loads settings from a root file if available and uses that as defaults in
        place of built in defaults.

        This allows setting of the config file path through environment variables.
        """
        from monty.serialization import loadfn

        config_file_path: str = values.get("CONFIG_FILE", DEFAULT_CONFIG_FILE_PATH)

        new_values = {}
        if Path(config_file_path).exists():
            new_values.update(loadfn(config_file_path))

        store = new_values.get("JOB_STORE")
        if isinstance(store, str):
            new_values["JOB_STORE"] = JobStore.from_file(store)
        elif isinstance(store, dict) and store.get("@class") == "JobStore":
            new_values["JOB_STORE"] = JobStore.from_dict(store)
        elif isinstance(store, dict):
            new_values["JOB_STORE"] = JobStore.from_dict_spec(store)

        new_values.update(values)
        return new_values



def test_settings_init():
    import os

    from maggma.stores import MemoryStore

    # set the config file to a not existing path so that it does not
    # pick the local configuration
    os.environ["JOBFLOW_CONFIG_FILE"] = "/some/not/existing/path"

    from jobflow import SETTINGS

    # assert default job store initialised correctly
    assert isinstance(SETTINGS.JOB_STORE.docs_store, MemoryStore)
    assert len(SETTINGS.JOB_STORE.additional_stores) == 0
    data_store = SETTINGS.JOB_STORE.additional_stores["data"]
    assert len(SETTINGS.JOB_STORE.additional_stores) == 1
    assert isinstance(data_store, MemoryStore)
    assert data_store is SETTINGS.JOB_STORE.additional_stores["data"]


def test_settings_object(clean_dir, test_data):
    import os
    from pathlib import Path

    from monty.serialization import dumpfn

    from jobflow import JobStore
    from jobflow.settings import JobflowSettings

    monty_spec = {
        "@module": "jobflow.core.store",
        "@class": "JobStore",
        "@version": "",
        "docs_store": {
            "@module": "maggma.stores.mongolike",
            "@class": "MemoryStore",
            "@version": "0.31.0",
            "collection_name": "memory_db_123",
        },
        "additional_stores": {},
        "save": {},
        "load": False,
    }

    dict_spec = {
        "docs_store": {
            "type": "MongoStore",
            "database": "jobflow_unittest",
            "collection_name": "outputs_567",
            "host": "localhost",
            "port": 27017,
        }
    }

    s3_store_spec = {
        "docs_store": {
            "type": "MemoryStore",
            "collection_name": "docs_store_123",
        },
        "additional_stores": {
            "data": {
                "type": "S3Store",
                "bucket": "bucket_123",
                "index": {
                    "type": "MemoryStore",
                },
            }
        },
    }

    # set the path to lood settings from
    os.environ["JOBFLOW_CONFIG_FILE"] = str(Path.cwd() / "config.yaml")

    # assert loading monty spec from files works
    dumpfn({"JOB_STORE": monty_spec}, "config.yaml")
    settings = JobflowSettings()
    assert settings.JOB_STORE.docs_store.collection_name == "memory_db_123"

    # assert loading alternative dict spec from files works
    dumpfn({"JOB_STORE": dict_spec}, "config.yaml")
    settings = JobflowSettings()
    assert settings.JOB_STORE.docs_store.collection_name == "outputs_567"

    dumpfn({"JOB_STORE": s3_store_spec}, "config.yaml")
    settings = JobflowSettings()
    assert settings.JOB_STORE.additional_stores["data"].bucket == "bucket_123"

    # assert loading from db file works.
    dumpfn({"JOB_STORE": str(test_data / "db.yaml")}, "config.yaml")
    settings = JobflowSettings()
    assert settings.JOB_STORE.docs_store.collection_name == "outputs"

    # assert loading from serialized file works.
    dumpfn({"JOB_STORE": str(test_data / "db_serialized.json")}, "config.yaml")
    settings = JobflowSettings()
    assert settings.JOB_STORE.docs_store.database == "jobflow_unittest"

    # assert passing a jobflow object works
    settings = JobflowSettings(JOB_STORE=JobStore.from_dict(monty_spec))
    assert settings.JOB_STORE.docs_store.collection_name == "memory_db_123"



"""Tools for constructing Job and Flow graphs."""

from __future__ import annotations

import warnings

import networkx as nx
from monty.dev import requires

try:
    import matplotlib
except ImportError:
    matplotlib = None

import typing

if typing.TYPE_CHECKING:
    pass

    import jobflow

__all__ = ["itergraph", "draw_graph", "to_pydot"]


def itergraph(graph: nx.DiGraph):
    """
    Iterate through a graph using a topological sort order.

    This means the nodes are yielded such that for every directed edge (u v)
    node u comes before v in the ordering.

    Parameters
    ----------
    graph
        A networkx graph.

    Raises
    ------
    ValueError
        If the graph contains cycles.

    Yields
    ------
    str
        The node uuid.
    """
    if not nx.is_directed_acyclic_graph(graph):
        raise ValueError("Graph is not acyclic, cannot determine dependency order.")

    subgraphs = [graph.subgraph(c) for c in nx.weakly_connected_components(graph)]

    if len(subgraphs) > 1:
        warnings.warn("Some jobs are not connected, their ordering may be random")

    for subgraph in subgraphs:
        yield from nx.topological_sort(subgraph)


@requires(matplotlib, "matplotlib must be installed to plot flow graphs.")
def draw_graph(
    graph: nx.DiGraph,
    layout_function: typing.Callable = None,
    figsize: tuple[float, float] = (12, 8),
):
    """
    Draw a networkx graph.

    Parameters
    ----------
    graph
        A graph object.
    layout_function
        A networkx layout function to use as the graph layout. For example,
        :obj:`.planar_layout`.
    figsize
        The figure size as a tuple of ``(width, height)``.

    Returns
    -------
    matplotlib.pyplot
        The matplotlib pyplot object.
    """
    import matplotlib.pyplot as plt

    if layout_function is None:
        try:
            pos = nx.nx_pydot.graphviz_layout(graph, prog="dot")
        except (ImportError, FileNotFoundError):
            # graphviz not installed
            pos = nx.planar_layout(graph)
    else:
        pos = layout_function(graph)

    plt.figure(figsize=figsize)

    nodes = graph.nodes()
    labels = nx.get_node_attributes(graph, "label")

    nx.draw_networkx_edges(graph, pos)
    nx.draw_networkx_nodes(
        graph, pos, nodelist=nodes, node_color="#B65555", linewidths=1, edgecolors="k"
    )
    nx.draw_networkx_labels(graph, pos, labels=labels)

    edge_labels = nx.get_edge_attributes(graph, "properties")
    nx.draw_networkx_edge_labels(graph, pos, edge_labels=edge_labels, rotate=False)

    plt.gca().axis("off")

    return plt


def to_pydot(flow: jobflow.Flow):
    """
    Convert a flow to a pydot graph.

    Pydot graphs can be visualised using graphviz and support more advanced features
    than networkx graphs. For example, the pydot graph also includes the flow
    containers.

    .. Note::
        Requires pydot and graphviz to be installed.

    Parameters
    ----------
    flow
        A flow.

    Returns
    -------
    pydot.Dot
        The pydot graph.

    Examples
    --------
    The pydot graph can be generated from a flow using:

    >>> from jobflow import job, Flow
    >>> @job
    ... def add(a, b):
    ...     return a + b
    >>> add_first = add(1, 2)
    >>> add_second = add(add_first.output, 2)
    >>> my_flow = Flow(jobs=[add_first, add_second])
    >>> graph = to_pydot(my_flow)

    If graphviz is installed, the pydot graph can be rendered to a file using:

    >>> graph.write("output.png", format="png")
    """
    import pydot

    from jobflow import Flow

    nx_graph = flow.graph
    pydot_graph = pydot.Dot(f'"{flow.name}"', graph_type="digraph")

    for n, nodedata in nx_graph.nodes(data=True):
        str_nodedata = {k: str(v) for k, v in nodedata.items()}
        p = pydot.Node(str(n), **str_nodedata)
        pydot_graph.add_node(p)

    for u, v, edgedata in nx_graph.edges(data=True):
        str_edgedata = {k: str(v) for k, v in edgedata.items()}
        edge = pydot.Edge(str(u), str(v), label=str_edgedata["properties"])
        pydot_graph.add_edge(edge)

    def add_cluster(nested_flow, outer_graph):
        cluster = pydot.Cluster(nested_flow.uuid)
        cluster.set_label(nested_flow.name)
        for job in nested_flow.jobs:
            if isinstance(job, Flow):
                add_cluster(job, cluster)
            else:
                cluster.add_node(pydot_graph.get_node(f'"{job.uuid}"')[0])

        outer_graph.add_subgraph(cluster)

    add_cluster(flow, pydot_graph)

    return pydot_graph



"""Define the primary jobflow database interface."""

from __future__ import annotations

import typing

from maggma.core import Store
from monty.json import MSONable

from jobflow.core.reference import OnMissing

if typing.TYPE_CHECKING:
    from enum import Enum
    from pathlib import Path
    from typing import Any, Dict, Iterator, List, Optional, Type, Union

    from maggma.core import Sort

    obj_type = Union[str, Enum, Type[MSONable], List[Union[Enum, str, Type[MSONable]]]]
    save_type = Optional[Dict[str, obj_type]]
    load_type = Union[bool, Dict[str, Union[bool, obj_type]]]

__all__ = ["JobStore"]

T = typing.TypeVar("T", bound="JobStore")


class JobStore(Store):
    """Store intended to allow pushing and pulling documents into multiple stores.

    Parameters
    ----------
    docs_store
        Store for basic documents.
    additional_stores
        Additional stores to use for storing large data/specific objects. Given as a
        mapping of ``{store name: store}`` where store is a maggma :obj:`.Store` object.
    save
        Which items to save in additional stores when uploading documents. Given as a
        mapping of ``{store name: store_type}`` where ``store_type`` can be a dictionary
        key (string or enum), an :obj:`.MSONable` class, or a list of keys/classes.
    load
        Which items to load from additional stores when querying documents. Given as a
        mapping of ``{store name: store_type}`` where ``store_type`` can be `True``, in
        which case all saved items are loaded, a dictionary key (string or enum),
        an :obj:`.MSONable` class, or a list of keys/classes. Alternatively,
        ``load=True`` will automatically load all items from every additional store.
    """

    def __init__(
        self,
        docs_store: Store,
        additional_stores: dict[str, Store] | None = None,
        save: save_type = None,
        load: load_type = False,
    ):
        self.docs_store = docs_store
        if additional_stores is None:
            self.additional_stores = {}
        else:
            self.additional_stores = additional_stores

        # enforce uuid key
        self.docs_store.key = "uuid"
        for additional_store in self.additional_stores.values():
            additional_store.key = "blob_uuid"
            if hasattr(additional_store, "searchable_fields"):
                additional_store.searchable_fields.extend(["job_uuid", "job_index"])

        if save is None or save is False:
            save = {}
        self.save = save

        if load is None:
            load = False
        self.load = load

        kwargs = {
            k: getattr(docs_store, k)
            for k in ("key", "last_updated_field", "last_updated_type")
        }
        super().__init__(**kwargs)

    @property
    def _collection(self):
        """Get the underlying collection object."""
        return self.docs_store._collection

    @property
    def name(self) -> str:
        """Get the name of the data source.

        Returns
        -------
        str
            A string representing this data source.
        """
        return f"JobStore-{self.docs_store.name}"

    def connect(self, force_reset: bool = False):
        """
        Connect to the source data.

        Parameters
        ----------
        force_reset
            Whether to reset the connection or not.
        """
        self.docs_store.connect(force_reset=force_reset)
        for additional_store in self.additional_stores.values():
            additional_store.connect(force_reset=force_reset)

    def close(self):
        """Close any connections."""
        self.docs_store.close()
        for additional_store in self.additional_stores.values():
            additional_store.close()

    def count(self, criteria: dict | None = None) -> int:
        """
        Count the number of documents matching the query criteria.

        Parameters
        ----------
        criteria
            PyMongo filter for documents to count in.

        Returns
        -------
        int
            The number of documents matching the query.
        """
        return self.docs_store.count(criteria=criteria)

    def query(
        self,
        criteria: dict | None = None,
        properties: dict | list | None = None,
        sort: dict[str, Sort | int] | None = None,
        skip: int = 0,
        limit: int = 0,
        load: load_type = None,
    ) -> Iterator[dict]:
        """
        Query the JobStore for documents.

        Parameters
        ----------
        criteria
            PyMongo filter for documents to search in.
        properties
            Properties to return in grouped documents.
        sort
            Dictionary of sort order for fields. Keys are field names and values are 1
            for ascending or -1 for descending.
        skip
            Number of documents to skip.
        limit
            Limit on the total number of documents returned.
        load
            Which items to load from additional stores. See ``JobStore`` constructor for
            more details.

        Yields
        ------
        Dict
            The documents.
        """
        from pydash import get

        from jobflow.utils.find import find_key, update_in_dictionary

        if load is None:
            load = self.load

        load_keys = _prepare_load(load)

        if isinstance(properties, (list, tuple)):
            properties += ["uuid", "index"]
        elif isinstance(properties, dict):
            properties.update({"uuid": 1, "index": 1})

        docs = self.docs_store.query(
            criteria=criteria, properties=properties, sort=sort, skip=skip, limit=limit
        )

        for doc in docs:
            if load_keys:
                # Process is
                # 1. Find the locations of all blob identifiers.
                # 2. Filter the locations based on the load criteria.
                # 3. Resolve all data blobs using the data store.
                # 4. Insert the data blobs into the document.
                locations = find_key(doc, "blob_uuid")
                all_blobs = [get(doc, list(loc)) for loc in locations]
                grouped_blobs = _filter_blobs(all_blobs, locations, load_keys)
                to_insert = {}

                for store_name, (blobs, locs) in grouped_blobs.items():
                    store = self.additional_stores.get(store_name, None)

                    if store is None:
                        raise ValueError(
                            f"Unrecognised additional store name: {store_name}"
                        )

                    object_info = {b["blob_uuid"]: loc for b, loc in zip(blobs, locs)}
                    objects = store.query(
                        criteria={"blob_uuid": {"$in": list(object_info.keys())}},
                        properties=["blob_uuid", "data"],
                    )
                    object_map = {o["blob_uuid"]: o["data"] for o in objects}
                    inserts = {tuple(v): object_map[k] for k, v in object_info.items()}
                    to_insert.update(inserts)

                update_in_dictionary(doc, to_insert)

            yield doc

    def query_one(
        self,
        criteria: dict | None = None,
        properties: dict | list | None = None,
        sort: dict[str, Sort | int] | None = None,
        load: load_type = None,
    ) -> dict | None:
        """
        Query the Store for a single document.

        Parameters
        ----------
        criteria
            PyMongo filter for documents to search.
        properties
            Properties to return in the document.
        sort
            Dictionary of sort order for fields. Keys are field names and values are 1
            for ascending or -1 for descending.
        load
            Which items to load from additional stores. See ``JobStore`` constructor for
            more details.

        Returns
        -------
        dict or None
            The document.
        """
        docs = self.query(
            criteria=criteria, properties=properties, load=load, sort=sort, limit=1
        )
        d = next(docs, None)
        return d

    def update(
        self,
        docs: list[dict] | dict,
        key: list | str | None = None,
        save: bool | save_type = None,
    ):
        """
        Update or insert documents into the Store.

        Parameters
        ----------
        docs
            The document or list of documents to update.
        key
            Field name(s) to determine uniqueness for a document, can be a list of
            multiple fields, a single field, or None if the Store's key field is to
            be used.
        save
            Which items to save in additional stores. See ``JobStore`` constructor for
            more details.
        """
        from collections import defaultdict
        from copy import deepcopy

        from monty.json import jsanitize
        from pydash import get

        from jobflow.utils.find import find_key, update_in_dictionary

        if save is None or save is True:
            save = self.save

        save_keys = _prepare_save(save)

        if not isinstance(docs, list):
            docs = [docs]

        if key is None:
            key = ["uuid", "index"]

        blob_data = defaultdict(list)
        dict_docs = []
        for doc in docs:
            doc = jsanitize(doc, strict=True, allow_bson=True)
            dict_docs.append(doc)

            if save_keys:
                locations = []
                for store_name, store_save in save_keys.items():
                    for save_key in store_save:
                        locations.extend(find_key(doc, save_key, include_end=True))

                    objects = [get(doc, list(loc)) for loc in locations]
                    object_map = {
                        tuple(k): o for k, o in zip(locations, objects) if o is not None
                    }
                    object_info = {
                        k: _get_blob_info(o, store_name) for k, o in object_map.items()
                    }
                    update_in_dictionary(doc, object_info)

                    # Now format blob data for saving in the data_store
                    for loc, data in object_map.items():
                        blob = deepcopy(object_info[loc])
                        blob.pop("store")
                        blob.update(
                            {
                                "data": data,
                                "job_uuid": doc["uuid"],
                                "job_index": doc["index"],
                            }
                        )
                        blob_data[store_name].append(blob)

        self.docs_store.update(dict_docs, key=key)

        for store_name, blobs in blob_data.items():
            # Here we use a try/except with a self.additional_stores[store_name]
            # instead of self.additional_store.get(store_name, None) to allow
            # for additional_stores to be a defaultdict with a MemoryStore's default.
            try:
                store = self.additional_stores[store_name]
            except KeyError:
                store = None

            if store is None:
                raise ValueError(f"Unrecognised additional store name: {store_name}")

            store.update(blobs, key="blob_uuid")

    def ensure_index(self, key: str, unique: bool = False) -> bool:
        """
        Try to create an index on document store and return True success.

        Parameters
        ----------
        key
            Single key to index.
        unique
            Whether or not this index contains only unique keys.

        Returns
        -------
        bool
            Whether the index exists/was created correctly.
        """
        return self.docs_store.ensure_index(key, unique=unique)

    def groupby(
        self,
        keys: list[str] | str,
        criteria: dict | None = None,
        properties: dict | list | None = None,
        sort: dict[str, Sort | int] | None = None,
        skip: int = 0,
        limit: int = 0,
        load: load_type = None,
    ) -> Iterator[tuple[dict, list[dict]]]:
        """
        Group documents by keys.

        Parameters
        ----------
        keys
            Fields to group documents.
        criteria
            PyMongo filter for documents to search in.
        properties
            Properties to return in grouped documents
        sort
            Dictionary of sort order for fields. Keys are field names and values are 1
            for ascending or -1 for descending.
        skip
            Number of documents to skip.
        limit
            Limit on the total number of documents returned.
        load
            Which items to load from additional stores. See ``JobStore`` constructor for
            more details.

        Yields
        ------
        dict, list[dict]
            The documents as (key, documents) grouped by their keys.
        """
        from itertools import groupby

        from pydash import get, has, set_

        keys = keys if isinstance(keys, (list, tuple)) else [keys]

        if isinstance(properties, dict):
            # make sure all keys are in properties...
            properties.update(dict(zip(keys, [1] * len(keys))))
        elif properties is not None:
            properties.extend(keys)

        docs = self.query(
            properties=properties,
            criteria=criteria,
            sort=sort,
            skip=skip,
            limit=limit,
            load=load,
        )
        data = [doc for doc in docs if all(has(doc, k) for k in keys)]

        def grouping_keys(doc):
            return tuple(get(doc, k) for k in keys)

        for vals, group in groupby(sorted(data, key=grouping_keys), key=grouping_keys):
            doc: dict[str, Any] = {}
            for k, v in zip(keys, vals):
                set_(doc, k, v)
            yield doc, list(group)

    def remove_docs(self, criteria: dict):
        """
        Remove docs matching the criteria.

        Parameters
        ----------
        criteria
            Criteria for documents to remove.
        """
        docs = self.query(criteria, properties=["uuid", "index"])
        for doc in docs:
            for store in self.additional_stores.values():
                store.remove_docs({"job_uuid": doc["uuid"], "job_index": doc["index"]})
        self.docs_store.remove_docs(criteria)

    def get_output(
        self,
        uuid: str,
        which: str | int = "last",
        load: load_type = False,
        cache: dict[str, Any] | None = None,
        on_missing: OnMissing = OnMissing.ERROR,
    ):
        """
        Get the output of a job UUID.

        Note that, unlike :obj:`JobStore.query`, this function will automatically
        try to resolve any output references in the job outputs.

        Parameters
        ----------
        uuid
            A job UUID.
        which
            If there are multiple job runs, which index to use. Options are:

            - ``"last"`` (default): Use the last job that ran.
            - ``"first"``: Use the first job that ran.
            - ``"all"``: Return all outputs, sorted with the lowest index first.
            - Alternatively, if an integer is specified, the output for this specific
              index will be queried.

        load
            Which items to load from additional stores. Setting to ``True`` will load
            all items stored in additional stores. See the ``JobStore`` constructor for
            more details.
        cache
            A dictionary cache to use for resolving of reference values.
        on_missing
            What to do if the output contains a reference and the reference cannot
            be resolved.

        Returns
        -------
        Any
            The output(s) for the job UUID.
        """
        from jobflow.core.reference import (
            find_and_get_references,
            find_and_resolve_references,
        )

        # TODO: Currently, OnMissing is not respected if a reference cycle is detected
        #       this could be fixed but will require more complicated logic just to
        #       catch a very unlikely event.

        if isinstance(which, int) or which in ("last", "first"):
            sort = -1 if which == "last" else 1

            criteria: dict[str, Any] = {"uuid": uuid}
            if isinstance(which, int):
                # handle specific index specified
                criteria["index"] = which

            result = self.query_one(
                criteria=criteria,
                properties=["output"],
                sort={"index": sort},
                load=load,
            )

            if result is None:
                istr = f" (index: {which})" if isinstance(which, int) else ""
                raise ValueError(f"UUID: {uuid}{istr} has no outputs.")

            refs = find_and_get_references(result["output"])
            if any([ref.uuid == uuid for ref in refs]):
                raise RuntimeError("Reference cycle detected - aborting.")

            return find_and_resolve_references(
                result["output"], self, cache=cache, on_missing=on_missing
            )
        else:
            results = list(
                self.query(
                    criteria={"uuid": uuid},
                    properties=["output"],
                    sort={"index": 1},
                    load=load,
                )
            )

            if len(results) == 0:
                raise ValueError(f"UUID: {uuid} has no outputs.")

            results = [r["output"] for r in results]

            refs = find_and_get_references(results)
            if any([ref.uuid == uuid for ref in refs]):
                raise RuntimeError("Reference cycle detected - aborting.")

            return find_and_resolve_references(
                results, self, cache=cache, on_missing=on_missing
            )

    @classmethod
    def from_file(cls: type[T], db_file: str | Path, **kwargs) -> T:
        """
        Create an JobStore from a database file.

        Two options are supported for the database file. The file should be in json or
        yaml format.

        The simplest format is a monty dumped version of the store, generated using:

        >>> from monty.serialization import dumpfn
        >>> dumpfn("job_store.yaml", job_store)

        Alternatively, the file can contain the keys docs_store, additional_stores and
        any other keyword arguments supported by the :obj:`JobStore` constructor. The
        docs_store and additional stores are specified by the ``type`` key which must
        match a Maggma ``Store`` subclass, and the remaining keys are passed to
        the store constructor. For example, the following file would  create a
        :obj:`JobStore` with a ``MongoStore`` for docs and a ``GridFSStore`` as an
        additional store for data.

        .. code-block:: yaml

            docs_store:
              type: MongoStore
              database: jobflow_unittest
              collection_name: outputs
              host: localhost
              port: 27017
            additional_stores:
              data:
                type: GridFSStore
                database: jobflow_unittest
                collection_name: outputs_blobs
                host: localhost
                port: 27017

        Parameters
        ----------
        db_file
            Path to the file containing the credentials.
        **kwargs
            Additional keyword arguments that get passed to the JobStore constructor.
            These arguments are ignored if the file contains a monty serialised
            JobStore.

        Returns
        -------
        JobStore
            A JobStore.
        """
        from monty.serialization import loadfn

        store_info = loadfn(db_file)

        if isinstance(store_info, JobStore):
            return store_info

        return cls.from_dict_spec(store_info, **kwargs)

    @classmethod
    def from_dict_spec(cls: type[T], spec: dict, **kwargs) -> T:
        """
        Create an JobStore from a dict specification.

        .. note::
            This function is different to ``JobStore.from_dict`` which is used to load
            the monty serialised representation of the class.

        The dictionary should contain the keys "docs_store", "additional_stores" and
        any other keyword arguments supported by the :obj:`JobStore` constructor. The
        docs_store and additional stores are specified by the ``type`` key which must
        match a Maggma ``Store`` subclass, and the remaining keys are passed to
        the store constructor. For example, the following file would  create a
        :obj:`JobStore` with a ``MongoStore`` for docs and a ``GridFSStore`` as an
        additional store for data.

        .. code-block:: yaml

            docs_store:
              type: MongoStore
              database: jobflow_unittest
              collection_name: outputs
              host: localhost
              port: 27017
            additional_stores:
              data:
                type: GridFSStore
                database: jobflow_unittest
                collection_name: outputs_blobs
                host: localhost
                port: 27017

        Parameters
        ----------
        spec
            The dictionary specification.
        **kwargs
            Additional keyword arguments that get passed to the JobStore constructor.

        Returns
        -------
        JobStore
            A JobStore.
        """
        import maggma.stores  # required to enable subclass searching

        if "docs_store" not in spec:
            raise ValueError("Unrecognised database file format.")

        def all_subclasses(cl):
            return set(cl.__subclasses__()).union(
                [s for c in cl.__subclasses__() for s in all_subclasses(c)]
            )

        all_stores = {s.__name__: s for s in all_subclasses(maggma.stores.Store)}

        docs_store_info = spec["docs_store"]
        docs_store = _construct_store(docs_store_info, all_stores)

        additional_stores = {}
        if "additional_stores" in spec:
            for store_name, info in spec["additional_stores"].items():
                additional_stores[store_name] = _construct_store(info, all_stores)
        return cls(docs_store, additional_stores, **kwargs)


def _construct_store(spec_dict, valid_stores):
    """Parse the dict containing {"type": <StoreType>} recursively."""
    store_type = spec_dict.pop("type")
    for k, v in spec_dict.items():
        if isinstance(v, dict) and "type" in v:
            spec_dict[k] = _construct_store(v, valid_stores)
    return valid_stores[store_type](**spec_dict)


def _prepare_load(
    load: load_type,
) -> bool | dict[str, bool | list[str | tuple[str, str]]]:
    """Standardize load types."""
    from enum import Enum

    if isinstance(load, bool):
        return load

    new_load: dict[str, bool | list[str | tuple[str, str]]] = {}
    for store_name, store_load in load.items():
        if isinstance(store_load, bool):
            new_load[store_name] = store_load
        else:
            if not isinstance(store_load, (tuple, list)):
                store_load = [store_load]

            new_store_load = []
            for ltype in store_load:
                if isinstance(ltype, Enum):
                    new_store_load.append(ltype.value)
                elif not isinstance(ltype, str) and issubclass(ltype, MSONable):
                    new_store_load.append((ltype.__module__, ltype.__name__))
                else:
                    new_store_load.append(ltype)

            new_load[store_name] = new_store_load
    return new_load


def _prepare_save(
    save: bool | save_type,
) -> dict[str, list[str | type[MSONable]]]:
    """Standardize save type."""
    from enum import Enum

    if isinstance(save, bool):
        return {}

    new_save = {}
    for store_name, store_save in save.items():
        if not isinstance(store_save, (tuple, list, bool)):
            store_save = [store_save]

        new_save[store_name] = [
            o.value if isinstance(o, Enum) else o for o in store_save
        ]
    return new_save


def _filter_blobs(
    blob_infos: list[dict],
    locations: list[list[Any]],
    load: bool | dict[str, bool | list[str | tuple[str, str]]] = None,
) -> dict[str, tuple[list[dict], list[list[Any]]]]:
    """Filter and group blobs."""
    from collections import defaultdict

    def _group_blobs(infos, locs):
        grouped = defaultdict(lambda: ([], []))
        for info, loc in zip(infos, locs):
            grouped[info["store"]][0].append(info)
            grouped[info["store"]][1].append(loc)
        return grouped

    if isinstance(load, bool):
        # return all blobs (load has to be True to get here)
        return _group_blobs(blob_infos, locations)

    new_blobs = []
    new_locations = []
    for _store_name, store_load in load.items():
        for blob, location in zip(blob_infos, locations):
            if store_load is True:
                new_blobs.append(blob)
                new_locations.append(location)
            elif isinstance(store_load, bool):
                # could just write "elif store_load is False" but need this for mypy
                continue
            else:
                for ltype in store_load:
                    if (
                        isinstance(ltype, tuple)
                        and blob.get("@class", None) == ltype[1]
                        and blob.get("@module", None) == ltype[0]
                    ) or location[-1] == ltype:
                        new_blobs.append(blob)
                        new_locations.append(location)

    return _group_blobs(new_blobs, new_locations)


def _get_blob_info(obj: Any, store_name: str) -> dict[str, str]:
    from jobflow.utils.uuid import suuid

    class_name = ""
    module_name = ""
    if isinstance(obj, dict) and "@class" in obj and "@module" in obj:
        class_name = obj["@class"]
        module_name = obj["@module"]

    return {
        "@class": class_name,
        "@module": module_name,
        "blob_uuid": suuid(),
        "store": store_name,
    }



"""Functions and classes for representing Job objects."""

from __future__ import annotations

import logging
import typing
import warnings
from dataclasses import dataclass, field

from monty.json import MSONable, jsanitize

from jobflow.core.reference import OnMissing, OutputReference
from jobflow.utils.uuid import suuid

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Hashable

    from networkx import DiGraph
    from pydantic import BaseModel

    import jobflow

logger = logging.getLogger(__name__)

__all__ = ["job", "Job", "Response", "JobConfig", "store_inputs"]


@dataclass
class JobConfig(MSONable):
    """
    The configuration parameters for a job.

    Parameters
    ----------
    resolve_references
        Whether to resolve any references before the job function is executed.
        If ``False`` the unresolved reference objects will be passed into the function
        call.
    on_missing_references
        What to do if the references cannot be resolved. The default is to throw an
        error.
    manager_config
        The configuration settings to control the manager execution.
    expose_store
        Whether to expose the store in :obj:`.CURRENT_JOB`` when the job is running.
    pass_manager_config
        Whether to pass the manager configuration on to detour, addition, and
        replacement jobs.
    response_manager_config
        The custom configuration to pass to a detour, addition, or replacement job.
        Using this kwarg will automatically take precedence over the behavior of
        ``pass_manager_config`` such that a different configuration than
        ``manger_config`` can be passed to downstream jobs.

    Returns
    -------
    JobConfig
        A :obj:`JobConfig` object.
    """

    resolve_references: bool = True
    on_missing_references: OnMissing = OnMissing.ERROR
    manager_config: dict = field(default_factory=dict)
    expose_store: bool = False
    pass_manager_config: bool = True
    response_manager_config: dict = field(default_factory=dict)


def job(method: Callable | None = None, **job_kwargs):
    """
    Wrap a function to produce a :obj:`Job`.

    :obj:`Job` objects are delayed function calls that can be used in an
    :obj:`Flow`. A job is composed of the function name and source and any
    arguments for the function. This decorator makes it simple to create
    job objects directly from a function definition. See the examples for more details.

    Parameters
    ----------
    method
        A function to wrap. This should not be specified directly and is implied
        by the decorator.
    **job_kwargs
        Other keyword arguments that will get passed to the :obj:`Job` init method.

    Examples
    --------
    >>> @job
    ... def print_message():
    ...     print("I am a Job")
    >>> print_job = print_message()
    >>> type(print_job)
    <class 'jobflow.core.job.Job'>
    >>> print_job.function
    <function print_message at 0x7ff72bdf6af0>

    Jobs can have required and optional parameters.

    >>> @job
    ... def print_sum(a, b=0):
    ...     return print(a + b)
    >>> print_sum_job = print_sum(1, 2)
    >>> print_sum_job.function_args
    (1, )
    >>> print_sum_job.function_kwargs
    {"b": 2}

    If the function returns a value it can be referenced using the ``output``
    attribute of the job.

    >>> @job
    ... def add(a, b):
    ...     return a + b
    >>> add_task = add(1, 2)
    >>> add_task.output
    OutputReference('abeb6f48-9b34-4698-ab69-e4dc2127ebe9')

    .. Note::
        Because the task has not yet been run, the output value is an
        :obj:`OutputReference` object. References are automatically converted to their
        computed values (resolved) when the task runs.

    If a dictionary of values is returned, the values can be indexed in the usual
    way.

    >>> @job
    ... def compute(a, b):
    ...     return {"sum": a + b, "product": a * b}
    >>> compute_task = compute(1, 2)
    >>> compute_task.output["sum"]
    OutputReference('abeb6f48-9b34-4698-ab69-e4dc2127ebe9', 'sum')

    .. Warning::
        If an output is indexed incorrectly, for example by trying to access a key that
        doesn't exist, this error will only be raised when the Job is executed.

    Jobs can return :obj:`.Response` objects that control the flow execution flow.
    For example, to replace the current job with another job, ``replace`` can be used.

    >>> from jobflow import Response
    >>> @job
    ... def replace(a, b):
    ...     new_job = compute(a, b)
    ...     return Response(replace=new_job)

    By default, job outputs are stored in the :obj`.JobStore` ``docs_store``. However,
    the :obj:`.JobStore` `additional_stores`` can also be used for job outputs. The
    stores are specified as keyword arguments, where the argument name gives the store
    name and the argument value is the type of data/key to store in that store. More
    details on the accepted key types are given in the :obj:`Job` docstring. In the
    example below, the "graph" key is stored in an additional store named "graphs" and
    the "data" key is stored in an additional store called "large_data".

    >>> @job(large_data="data", graphs="graph")
    ... def compute(a, b):
    ...     return {"large_data": b, "graph": a }

    .. Note::
        Using additional stores requires the :obj:`.JobStore` to be configured with
        the required store names present. See the :obj:`.JobStore` docstring for more
        details.

    See Also
    --------
    Job, .Flow, .Response
    """

    def decorator(func):
        from functools import wraps

        # unwrap staticmethod or classmethod decorators
        desc = next(
            (desc for desc in (staticmethod, classmethod) if isinstance(func, desc)),
            None,
        )

        if desc:
            func = func.__func__

        @wraps(func)
        def get_job(*args, **kwargs) -> Job:

            f = func
            if len(args) > 0:
                # see if the first argument has a function with the same name as
                # this function
                met = getattr(args[0], func.__name__, None)
                if met:
                    # if so, check to see if that function ha been wrapped and
                    # whether the unwrapped function is the same as this function
                    wrap = getattr(met, "__func__", None)
                    if getattr(wrap, "original", None) is func:
                        # Ah ha. The function is a bound method.
                        f = met
                        args = args[1:]

            return Job(
                function=f,
                function_args=args,
                function_kwargs=kwargs,
                **job_kwargs,
            )

        get_job.original = func

        if desc:
            # rewrap staticmethod or classmethod decorators
            get_job = desc(get_job)

        return get_job

    # See if we're being called as @job or @job().
    if method is None:
        # We're called with parens.
        return decorator

    # We're called as @job without parens.
    return decorator(method)


class Job(MSONable):
    """
    A :obj:`Job` is a delayed function call that can be used in an :obj:`.Flow`.

    In general, one should not create :obj:`Job` objects directly but instead use
    the :obj:`job` decorator on a function. Any calls to a decorated function will
    return an :obj:`Job` object.

    Parameters
    ----------
    function
        A function. Can be a builtin function such as ``sum`` or any other function
        provided it can be imported. Class and static methods can also be used, provided
        the class is importable. Lastly, methods (functions bound to an instance of
        class) can be used, provided the class is :obj:`.MSONable`.
    function_args
        The positional arguments to the function call.
    function_kwargs
        The keyword arguments to the function call.
    output_schema
        A pydantic model that defines the schema of the output.
    uuid
        A unique identifier for the job. Generated automatically.
    index
        The index of the job (number of times the job has been replaced).
    name
        The name of the job. If not set it will be determined from ``function``.
    metadata
        A dictionary of information that will get stored alongside the job output.
    config
        The config setting for the job.
    hosts
        The list of UUIDs of the hosts containing the job. The object identified by one
        UUID of the list should be contained in objects identified by its subsequent
        elements.
    metadata_updates
        A list of updates for the metadata that will be applied to any Flow/Job
        generated by the job.
    config_updates
        A list of updates for the config that will be applied to any Flow/Job generated
        by the job.
    **kwargs
        Additional keyword arguments that can be used to specify which outputs to save
        in additional stores. The argument name gives the additional store name and the
        argument value gives the type of data to store in that additional store.
        The value can be ``True`` in which case all outputs are stored in the additional
        store, a dictionary key (string or enum), an :obj:`.MSONable` class, or a list
        of keys/classes.

    Attributes
    ----------
    output
        The output of the job. This is a reference to the future job output and
        can be used as the input to other Jobs or Flows.

    Returns
    -------
    Job
        A job.

    Examples
    --------
    Builtin functions such as :obj:`print` can be specified.

    >>> print_task = Job(function=print, args=("I am a job", ))

    Or other functions of the Python standard library.

    >>> import os
    >>> Job(function=os.path.join, args=("folder", "filename.txt"))

    To use custom functions, the functions should be importable (i.e. not
    defined in another function). For example, if the following function is defined
    in the ``my_package`` module.

    >>> def add(a, b):
    ...     return a + b
    >>> add_job = Job(function=add, args=(1, 2))

    More details are given in the :obj:`job` decorator docstring.

    See Also
    --------
    job, Response, .Flow
    """

    def __init__(
        self,
        function: Callable,
        function_args: tuple[Any, ...] = None,
        function_kwargs: dict[str, Any] = None,
        output_schema: type[BaseModel] | None = None,
        uuid: str = None,
        index: int = 1,
        name: str | None = None,
        metadata: dict[str, Any] = None,
        config: JobConfig = None,
        hosts: list[str] | None = None,
        metadata_updates: list[dict[str, Any]] | None = None,
        config_updates: list[dict[str, Any]] | None = None,
        **kwargs,
    ):
        from copy import deepcopy

        from jobflow.utils.find import contains_flow_or_job

        function_args = () if function_args is None else function_args
        function_kwargs = {} if function_kwargs is None else function_kwargs
        uuid = suuid() if uuid is None else uuid
        metadata = {} if metadata is None else metadata
        config = JobConfig() if config is None else config

        # make a deep copy of the function (means makers do not share the same instance)
        self.function = deepcopy(function)
        self.function_args = function_args
        self.function_kwargs = function_kwargs
        self.output_schema = output_schema
        self.uuid = uuid
        self.index = index
        self.name = name
        self.metadata = metadata
        self.config = config
        self.hosts = hosts or []
        self.metadata_updates = metadata_updates or []
        self.config_updates = config_updates or []
        self._kwargs = kwargs

        if sum(v is True for v in kwargs.values()) > 1:
            raise ValueError("Cannot select True for multiple additional stores.")

        if self.name is None:
            if self.maker is not None:
                self.name = self.maker.name
            else:
                self.name = getattr(function, "__qualname__", function.__name__)

        self.output = OutputReference(self.uuid, output_schema=self.output_schema)

        # check to see if job or flow is included in the job args
        # this is a possible situation but likely a mistake
        all_args = tuple(self.function_args) + tuple(self.function_kwargs.values())
        if contains_flow_or_job(all_args):
            warnings.warn(
                f"Job '{self.name}' contains an Flow or Job as an input. "
                f"Usually inputs should be the output of a Job or an Flow (e.g. "
                f"job.output). If this message is unexpected then double check the "
                f"inputs to your Job."
            )

    @property
    def input_references(self) -> tuple[jobflow.OutputReference, ...]:
        """
        Find :obj:`.OutputReference` objects in the job inputs.

        Returns
        -------
        tuple(OutputReference, ...)
            The references in the inputs to the job.
        """
        from jobflow.core.reference import find_and_get_references

        references: set[jobflow.OutputReference] = set()
        for arg in tuple(self.function_args) + tuple(self.function_kwargs.values()):
            references.update(find_and_get_references(arg))

        return tuple(references)

    @property
    def input_uuids(self) -> tuple[str, ...]:
        """
        Uuids of any :obj:`.OutputReference` objects in the job inputs.

        Returns
        -------
        tuple(str, ...)
           The UUIDs of the references in the job inputs.
        """
        return tuple(ref.uuid for ref in self.input_references)

    @property
    def input_references_grouped(self) -> dict[str, tuple[OutputReference, ...]]:
        """
        Group any :obj:`.OutputReference` objects in the job inputs by their UUIDs.

        Returns
        -------
        dict[str, tuple(OutputReference, ...)]
            The references grouped by their UUIDs.
        """
        from collections import defaultdict

        groups = defaultdict(set)
        for ref in self.input_references:
            groups[ref.uuid].add(ref)

        return {k: tuple(v) for k, v in groups.items()}

    @property
    def maker(self) -> jobflow.Maker | None:
        """
        Get the host :obj:`.Maker` object if a job is to run a maker function.

        Returns
        -------
        Maker or None
            A maker object.
        """
        from jobflow import Maker

        bound = getattr(self.function, "__self__", None)
        if isinstance(bound, Maker):
            return bound
        return None

    @property
    def graph(self) -> DiGraph:
        """
        Get a graph of the job indicating the inputs to the job.

        Returns
        -------
        DiGraph
            The graph showing the connectivity of the jobs.
        """
        from networkx import DiGraph

        edges = []
        for uuid, refs in self.input_references_grouped.items():
            properties: list[str] | str = [
                ref.attributes_formatted[-1]
                .replace("[", "")
                .replace("]", "")
                .replace(".", "")
                for ref in refs
                if ref.attributes
            ]
            properties = properties[0] if len(properties) == 1 else properties
            properties = properties if len(properties) > 0 else "output"
            edges.append((uuid, self.uuid, {"properties": properties}))

        graph = DiGraph()
        graph.add_node(self.uuid, job=self, label=self.name)
        graph.add_edges_from(edges)
        return graph

    @property
    def host(self):
        """
        UUID of the first Flow that contains the Job.

        Returns
        -------
        str
            the UUID of the host.
        """
        return self.hosts[0] if self.hosts else None

    def set_uuid(self, uuid: str):
        """
        Set the UUID of the job.

        Parameters
        ----------
        uuid
            A UUID.
        """
        self.uuid = uuid
        self.output = self.output.set_uuid(uuid)

    def run(self, store: jobflow.JobStore) -> Response:
        """
        Run the job.

        If the job has inputs that are :obj:`.OutputReference` objects, then they will
        need to be resolved before the job can run. See the docstring for
        :obj:`.OutputReference.resolve()` for more details.

        Parameters
        ----------
        store
            A :obj:`.JobStore` to use for resolving references and storing job outputs.

        Returns
        -------
        Response
            The response of the job, containing the outputs, and other settings that
            determine the flow execution.

        Raises
        ------
        ImportError
            If the job function cannot be imported.

        See Also
        --------
        Response, .OutputReference
        """
        import builtins
        import types
        from datetime import datetime

        from jobflow import CURRENT_JOB
        from jobflow.core.flow import get_flow

        index_str = f", {self.index}" if self.index != 1 else ""
        logger.info(f"Starting job - {self.name} ({self.uuid}{index_str})")
        CURRENT_JOB.job = self

        if self.config.expose_store:
            CURRENT_JOB.store = store

        if self.config.resolve_references:
            self.resolve_args(store=store)

        # if Job was created using the job decorator, then access the original function
        function = getattr(self.function, "original", self.function)

        # if function is bound method we need to do some magic to bind the unwrapped
        # function to the class/instance
        bound = getattr(self.function, "__self__", None)
        if bound is not None and bound is not builtins:
            function = types.MethodType(function, bound)

        response = function(*self.function_args, **self.function_kwargs)
        response = Response.from_job_returns(response, self.output_schema)

        if response.replace is not None:
            response.replace = prepare_replace(response.replace, self)

        if response.addition is not None:
            # wrap the detour in a Flow to avoid problems if it needs to get
            # wrapped at a later stage
            response.addition = get_flow(response.addition)

        if response.detour is not None:
            # wrap the detour in a Flow to avoid problems if it needs to get
            # wrapped at a later stage
            response.detour = get_flow(response.detour)

        # common actions that should be applied to all newly generated Jobs/Flows
        for new_jobs in (response.replace, response.addition, response.detour):
            if new_jobs is not None:
                new_jobs.add_hosts_uuids(self.hosts)
                for metadata_update in self.metadata_updates:
                    new_jobs.update_metadata(**metadata_update, dynamic=True)
                for config_update in self.config_updates:
                    new_jobs.update_config(**config_update, dynamic=True)

        if self.config.response_manager_config:
            passed_config = self.config.response_manager_config
        elif self.config.pass_manager_config:
            passed_config = self.config.manager_config
        else:
            passed_config = None

        if passed_config:

            if response.addition is not None:
                pass_manager_config(response.addition, passed_config)

            if response.detour is not None:
                pass_manager_config(response.detour, passed_config)

            if response.replace is not None:
                pass_manager_config(response.replace, passed_config)

        try:
            output = jsanitize(
                response.output, strict=True, enum_values=True, allow_bson=True
            )
        except AttributeError as err:
            raise RuntimeError(
                "Job output contained an object that is not MSONable and therefore "
                "could not be serialized."
            ) from err

        save = {k: "output" if v is True else v for k, v in self._kwargs.items()}
        data = {
            "uuid": self.uuid,
            "index": self.index,
            "output": output,
            "completed_at": datetime.now().isoformat(),
            "metadata": self.metadata,
            "hosts": self.hosts,
            "name": self.name,
        }
        store.update(data, key=["uuid", "index"], save=save)

        CURRENT_JOB.reset()
        logger.info(f"Finished job - {self.name} ({self.uuid}{index_str})")
        return response

    def resolve_args(
        self,
        store: jobflow.JobStore,
        inplace: bool = True,
    ) -> Job:
        """
        Resolve any :obj:`.OutputReference` objects in the input arguments.

        See the docstring for :obj:`.OutputReference.resolve()` for more details.

        Parameters
        ----------
        store
            A maggma store to use for resolving references.
        inplace
            Update the arguments of the current job or return a new job object.

        Returns
        -------
        Job
            A job with the references resolved.
        """
        from copy import deepcopy

        from jobflow.core.reference import find_and_resolve_references

        cache: dict[str, Any] = {}
        resolved_args = find_and_resolve_references(
            self.function_args,
            store,
            cache=cache,
            on_missing=self.config.on_missing_references,
        )
        resolved_kwargs = find_and_resolve_references(
            self.function_kwargs,
            store,
            cache=cache,
            on_missing=self.config.on_missing_references,
        )
        resolved_args = tuple(resolved_args)

        if inplace:
            self.function_args = resolved_args
            self.function_kwargs = resolved_kwargs
            return self

        new_job = deepcopy(self)
        new_job.function_args = resolved_args
        new_job.function_kwargs = resolved_kwargs
        return new_job

    def update_kwargs(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        function_filter: Callable | None = None,
        dict_mod: bool = False,
    ):
        """
        Update the kwargs of the jobs.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.

        Examples
        --------
        Consider a simple job with a ``number`` keyword argument.

        >>> from jobflow import job
        >>> @job
        ... def add(a, number=5):
        ...     return a + number
        >>> add_job = add(1)

        The ``number`` argument can be updated using.

        >>> add_job.update_kwargs({"number": 10})
        """
        from jobflow.utils.dict_mods import apply_mod

        if function_filter is not None and function_filter != self.function:
            return

        if (
            name_filter is not None
            and self.name is not None
            and name_filter not in self.name
        ):
            return

        # if we get to here then we pass all the filters
        if dict_mod:
            apply_mod(update, self.function_kwargs)
        else:
            self.function_kwargs.update(update)

    def update_maker_kwargs(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        class_filter: type[jobflow.Maker] | None = None,
        nested: bool = True,
        dict_mod: bool = False,
    ):
        """
        Update the keyword arguments of any :obj:`.Maker` objects in the job source.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the Maker name. Only Makers with a matching name will be
            updated. Includes partial matches, e.g. "ad" will match a Maker with the
            name "adder".
        class_filter
            A filter for the maker class. Only Makers with a matching class will be
            updated. Note the class filter will match any subclasses.
        nested
            Whether to apply the updates to Maker objects that are themselves kwargs
            of Maker, job, or flow objects. See examples for more details.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.

        Examples
        --------
        Consider the following job from a Maker:

        >>> from dataclasses import dataclass
        >>> from jobflow import job, Maker, Flow
        >>> @dataclass
        ... class AddMaker(Maker):
        ...     name: str = "add"
        ...     number: float = 10
        ...
        ...     @job
        ...     def make(self, a):
        ...         return a + self.number
        >>> maker = AddMaker()
        >>> add_job = maker.make(1)

        The ``number`` argument could be updated in the following ways.

        >>> add_job.update_maker_kwargs({"number": 10})

        By default, the updates are applied to nested Makers. These are Makers
        which are present in the kwargs of another Maker. Consider the following case
        for a Maker that produces a job that replaces itself with another job.

        >>> from jobflow import Response
        >>> @dataclass
        ... class ReplacementMaker(Maker):
        ...     name: str = "replace"
        ...     add_maker: Maker = AddMaker()
        ...
        ...     @job
        ...     def make(self, a):
        ...         add_job = self.add_maker.make(a)
        ...         return Response(replace=add_job)
        >>> maker = ReplacementMaker()
        >>> my_job = maker.make(1)

        The following update will apply to the nested ``AddMaker`` in the kwargs of the
        ``RestartMaker``:

        >>> my_job.update_maker_kwargs({"number": 10}, class_filter=AddMaker)

        However, if ``nested=False``, then the update will not be applied to the nested
        Maker:

        >>> my_job.update_maker_kwargs(
        ...     {"number": 10}, class_filter=AddMaker, nested=False
        ... )
        """
        from jobflow import Maker

        if self.maker is not None:
            maker = self.maker.update_kwargs(
                update,
                name_filter=name_filter,
                class_filter=class_filter,
                nested=nested,
                dict_mod=dict_mod,
            )
            self.function = getattr(maker, self.function.__name__)
        elif nested:
            # also look for makers in job args and kwargs
            new_args = []
            for arg in self.function_args:
                if isinstance(arg, Maker):
                    new_arg = arg.update_kwargs(
                        update,
                        name_filter=name_filter,
                        class_filter=class_filter,
                        nested=nested,
                        dict_mod=dict_mod,
                    )
                    new_args.append(new_arg)
                else:
                    new_args.append(arg)
            self.function_args = tuple(new_args)

            for name, kwarg in self.function_kwargs.items():
                if isinstance(kwarg, Maker):
                    self.function_kwargs[name] = kwarg.update_kwargs(
                        update,
                        name_filter=name_filter,
                        class_filter=class_filter,
                        nested=nested,
                        dict_mod=dict_mod,
                    )

    def append_name(self, append_str: str, prepend: bool = False):
        """
        Append a string to the name of the job.

        Parameters
        ----------
        append_str
            A string to append.
        prepend
            Prepend the name rather than appending it.
        """
        if prepend:
            self.name = append_str + self.name
        else:
            self.name += append_str

    def update_metadata(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        function_filter: Callable | None = None,
        dict_mod: bool = False,
        dynamic: bool = True,
    ):
        """
        Update the metadata of the job.

        Can optionally apply the same updates at runtime to any Job or Flow generated
        by this job.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.
        dynamic
            The updates will be propagated to Jobs/Flows dynamically generated at
            runtime.

        Examples
        --------
        Consider a simple job that makes use of a :obj:`Maker` to generate additional
        jobs at runtime (see :obj:`Response` options for more details):

        >>> @job
        ... def use_maker(maker):
        ...     return Response(replace=maker.make())

        Calling `update_metadata` with `dynamic` set to `True` (the default)

        >>> test_job = use_maker(ExampleMaker())
        ... test_job.update_metadata({"example": 1}, dynamic=True)

        will not only set the `example` metadata to the `test_job`, but also to all the
        new Jobs that will be generated at runtime by the ExampleMaker.

        `update_metadata` can be called multiple times with different `name_filter` or
        `function_filter` to control which Jobs will be updated.

        At variance, if `dynamic` is set to `False` the `example` metadata will only be
        added to the `test_job` and not to the generated Jobs.
        """
        from jobflow.utils.dict_mods import apply_mod

        if dynamic:
            dict_input = {
                "update": update,
                "name_filter": name_filter,
                "function_filter": function_filter,
                "dict_mod": dict_mod,
            }
            self.metadata_updates.append(dict_input)

        # unwrap the functions in case the job is a decorated one
        function_filter = getattr(function_filter, "__wrapped__", function_filter)
        function = getattr(self.function, "__wrapped__", self.function)

        # if function_filter is not None and function_filter != self.function:
        if function_filter is not None and function_filter != function:
            return

        if name_filter is not None and (
            self.name is None or name_filter not in self.name
        ):
            return

        # if we get to here then we pass all the filters
        if dict_mod:
            apply_mod(update, self.metadata)
        else:
            self.metadata.update(update)

    def update_config(
        self,
        config: JobConfig | dict,
        name_filter: str = None,
        function_filter: Callable = None,
        attributes: list[str] | str = None,
        dynamic: bool = True,
    ):
        """
        Update the job config.

        Can optionally apply the same updates at runtime to any Job or Flow generated
        by this job.

        Parameters
        ----------
        config
            A JobConfig object or a dict with containing the attributes to update.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        attributes :
            Which attributes of the job config to set. Can be specified as one or more
            attributes specified by their name.
        dynamic
            The updates will be propagated to Jobs/Flows dynamically generated at
            runtime.

        Examples
        --------
        Consider a simple job.

        >>> from jobflow import job, JobConfig
        >>> @job
        ... def add(a, b):
        ...     return a + b
        >>> add_job = add(1, 2)

        The ``config`` can be updated using.

        >>> new_config = JobConfig(
        ...    manager_config={"_fworker": "myfworker"}, resolve_references=False
        ... )
        >>> add_job.update_config(new_config)

        To only update specific attributes, the ``attributes`` argument can be
        specified. For example, the following will only update the "manager_config"
        attribute of the job config.

        >>> add_job.update_config(new_config, attributes="manager_config")

        Alternatively, the config can be specified as a dictionary with keys that are
        attributes of the JobConfig object. This allows you to specify updates without
        having to create a completely new JobConfig object. For example:

        >>> add_job.update_config({"manager_config": {"_fworker": "myfworker"}})

        Consider instead a simple job that makes use of a :obj:`Maker` to generate
        additional jobs at runtime (see :obj:`Response` options for more details):

        >>> @job
        ... def use_maker(maker):
        ...     return Response(replace=maker.make())

        Calling `update_config` with `dynamic` set to `True` (the default)

        >>> test_job = use_maker(ExampleMaker())
        ... test_job.update_config({"manager_config": {"_fworker": "myfworker"}})

        will not only set the `manager_config` to the `test_job`, but also to all the
        new Jobs that will be generated at runtime by the ExampleMaker.

        `update_config` can be called multiple times with different `name_filter` or
        `function_filter` to control which Jobs will be updated.

        At variance, if `dynamic` is set to `False` the `manager_config` option will
        only be set for the `test_job` and not for the generated Jobs.
        """
        if dynamic:
            dict_input = {
                "config": config,
                "name_filter": name_filter,
                "function_filter": function_filter,
                "attributes": attributes,
            }
            self.config_updates.append(dict_input)

        # unwrap the functions in case the job is a decorated one
        function_filter = getattr(function_filter, "__wrapped__", function_filter)
        function = getattr(self.function, "__wrapped__", self.function)

        # if function_filter is not None and function_filter != self.function:
        if function_filter is not None and function_filter != function:
            return

        if name_filter is not None and (
            self.name is None or name_filter not in self.name
        ):
            return

        # if we get to here then we pass all the filters
        if isinstance(config, dict):
            # convert dict specification to a JobConfig but set the attributes
            if attributes is None:
                attributes = list(config.keys())

            attributes = [attributes] if isinstance(attributes, str) else attributes
            if not set(attributes).issubset(set(config.keys())):
                raise ValueError(
                    "Specified attributes include a key that is not present in the "
                    "config dictionary."
                )
            config = JobConfig(**config)

        if attributes is None:
            # overwrite the whole config
            self.config = config
        else:
            # only update the specified attributes
            attributes = [attributes] if isinstance(attributes, str) else attributes
            for attr in attributes:
                if not hasattr(self.config, attr):
                    raise ValueError(f"Unknown JobConfig attribute: {attr}")
                setattr(self.config, attr, getattr(config, attr))

    def as_dict(self) -> dict:
        """Serialize the job as a dictionary."""
        d = super().as_dict()

        # fireworks can't serialize functions and classes, so explicitly serialize to
        # the job recursively using monty to avoid issues
        return jsanitize(d, strict=True, enum_values=True)

    def __setattr__(self, key, value):
        """Handle setting attributes. Implements a special case for job name."""
        if key == "name" and value is not None and self.maker is not None:
            # have to be careful and also update the name of the bound maker
            # the ``value is not None`` in the if statement is needed otherwise the name
            # of the maker will get set to None during class init
            self.__dict__[key] = value
            self.maker.name = value
        else:
            super().__setattr__(key, value)

    def add_hosts_uuids(self, hosts_uuids: str | list[str], prepend: bool = False):
        """
        Add a list of UUIDs to the internal list of hosts.

        The elements of the list are supposed to be ordered in such a way that
        the object identified by one UUID of the list is contained in objects
        identified by its subsequent elements.

        Parameters
        ----------
        hosts_uuids
            A list of UUIDs to add.
        prepend
            Insert the UUIDs at the beginning of the list rather than extending it.
        """
        if not isinstance(hosts_uuids, (list, tuple)):
            hosts_uuids = [hosts_uuids]
        if prepend:
            self.hosts[0:0] = hosts_uuids
        else:
            self.hosts.extend(hosts_uuids)


# For type checking, the Response output type can be specified
# in a type hint via this type variable.
# For example, a signature `-> Response[int]` would require
# that the Response.output is an int.
T = typing.TypeVar("T")


@dataclass
class Response(typing.Generic[T]):
    """
    The :obj:`Response` contains the output, detours, and stop commands of a job.

    Parameters
    ----------
    output
        The job output.
    detour
        A flow or job to detour to.
    addition
        A flow or job to add to the current flow.
    replace
        A flow or job to replace the current job.
    stored_data
        Data to be stored by the flow manager.
    stop_children
        Stop any children of the current flow.
    stop_jobflow
        Stop executing all remaining jobs.
    """

    output: T | None = None
    detour: jobflow.Flow | Job | list[Job] | list[jobflow.Flow] | None = None
    addition: jobflow.Flow | Job | list[Job] | list[jobflow.Flow] | None = None
    replace: jobflow.Flow | Job | list[Job] | list[jobflow.Flow] | None = None
    stored_data: dict[Hashable, Any] | None = None
    stop_children: bool = False
    stop_jobflow: bool = False

    @classmethod
    def from_job_returns(
        cls,
        job_returns: Any | None,
        output_schema: type[BaseModel] | None = None,
    ) -> Response:
        """
        Generate a :obj:`Response` from the outputs of a :obj:`Job`.

        Parameters
        ----------
        job_returns
            The outputs of a job. If this is a :obj:`Response` object, the output schema
            will be applied to the response outputs and the response returned.
            Otherwise, the ``job_returns`` will be put into the ``outputs`` of a new
            :obj:`Response` object.
        output_schema
            A pydantic model associated with the job. Used to enforce a schema for the
            outputs.

        Raises
        ------
        ValueError
            If the job outputs do not match the output schema.

        Returns
        -------
        Response
            The job response controlling the data to store and flow execution options.
        """
        if isinstance(job_returns, Response):
            if job_returns.replace is None:
                # only apply output schema if there is no replace.
                job_returns.output = apply_schema(job_returns.output, output_schema)

            return job_returns

        if isinstance(job_returns, (list, tuple)):
            # check that a Response object is not given as one of many outputs
            for r in job_returns:
                if isinstance(r, Response):
                    raise ValueError(
                        "Response cannot be returned in combination with other outputs."
                    )

        return cls(output=apply_schema(job_returns, output_schema))


def apply_schema(output: Any, schema: type[BaseModel] | None):
    """
    Apply schema to job outputs.

    Parameters
    ----------
    output
        The job outputs.
    schema
        A pydantic model that defines the schema to apply.

    Raises
    ------
    ValueError
        If a schema is set but there are no outputs.
    ValueError
        If the outputs do not match the schema.

    Returns
    -------
    BaseModel or Any
        Returns an instance of the schema if the schema is set or the original output.
    """
    if schema is None or isinstance(output, schema):
        return output

    if output is None:
        raise ValueError(f"Expected output of type {schema.__name__} but got no output")

    if not isinstance(output, dict):
        raise ValueError(
            f"Expected output to be {schema.__name__} or dict but got output type "
            f"of {type(output).__name__}."
        )

    return schema(**output)


@job(config=JobConfig(resolve_references=False, on_missing_references=OnMissing.NONE))
def store_inputs(inputs: Any) -> Any:
    """
    Job to store inputs.

    Note that any :obj:`Reference` objects will not be resolved, however, missing
    references will be replaced with ``None``.

    Parameters
    ----------
    inputs:
        The inputs to store.
    """
    return inputs


def prepare_replace(
    replace: jobflow.Flow | Job | list[Job],
    current_job: Job,
) -> jobflow.Flow:
    """
    Prepare a replacement :obj:`Flow` or :obj:`Job`.

    If the replacement is a ``Flow``, then an additional ``Job`` will be inserted
    that maps the output id of the original job to outputs of the ``Flow``.

    If the replacement is a ``Flow`` or a ``Job``, then this function pass on
    the manager config, schema, and metadata and set the according UUIDs and job index.

    Parameters
    ----------
    replace
        A :obj:`Flow` or :obj:`Job` to use as the replacement.
    current_job
        The current job.

    Returns
    -------
    Flow
        The updated flow.
    """
    from jobflow.core.flow import Flow

    if isinstance(replace, (list, tuple)):
        replace = Flow(jobs=replace)

    if isinstance(replace, Flow) and replace.output is not None:
        # add a job with same UUID as the current job to store the outputs of the
        # flow; this job will inherit the metadata and output schema of the current
        # job
        store_output_job = store_inputs(replace.output)
        store_output_job.set_uuid(current_job.uuid)
        store_output_job.index = current_job.index + 1
        store_output_job.metadata = current_job.metadata
        store_output_job.output_schema = current_job.output_schema
        replace.add_jobs(store_output_job)

    elif isinstance(replace, Job):
        # replace is a single Job
        replace.set_uuid(current_job.uuid)
        replace.index = current_job.index + 1

        metadata = replace.metadata
        metadata.update(current_job.metadata)
        replace.metadata = metadata

        if not replace.output_schema:
            replace.output_schema = current_job.output_schema

        replace = Flow(jobs=replace, output=replace.output)

    return replace


def pass_manager_config(
    jobs: Job | jobflow.Flow | list[Job | jobflow.Flow],
    manager_config: dict[str, Any],
):
    """
    Pass the manager config on to any jobs in the jobs array.

    Parameters
    ----------
    jobs
        A job, flow, or list of jobs/flows.
    manager_config
        A manager config to pass on.
    metadata
        Metadata to pass on.
    """
    from copy import deepcopy

    all_jobs: list[Job] = []

    def get_jobs(arg):

        if isinstance(arg, Job):
            all_jobs.append(arg)
        elif isinstance(arg, (list, tuple)):
            for j in arg:
                get_jobs(j)
        elif hasattr(arg, "jobs"):
            # this is a flow
            get_jobs(arg.jobs)
        else:
            raise ValueError("Unrecognised jobs format")

    # extract all jobs from the input array
    get_jobs(jobs)

    # update manager config
    for ajob in all_jobs:
        ajob.config.manager_config = deepcopy(manager_config)



"""Tools for generating UUIDs."""


def suuid() -> str:
    """
    Generate a string UUID (universally unique identifier).

    Uses the UUID4 specification.

    Returns
    -------
    str
        A UUID.
    """
    from uuid import uuid4

    return str(uuid4())



import pytest


@pytest.fixture
def uninstall_jobflow(monkeypatch):
    import pkg_resources

    original_func = pkg_resources.get_distribution

    def mock_func(name):
        if name == "jobflow":
            raise pkg_resources.DistributionNotFound()
        return original_func(name)

    monkeypatch.setattr(pkg_resources, "get_distribution", mock_func)


def test_installed_version():
    from jobflow import __version__

    assert __version__ != ""


@pytest.mark.usefixtures("uninstall_jobflow")
def test_not_installed_version(monkeypatch):
    import importlib

    from jobflow import _version

    # reimport modules
    importlib.reload(_version)

    # test version
    assert _version.__version__ == ""



"""Utility functions for logging, enumerations and manipulating dictionaries."""
from jobflow.utils.enum import ValueEnum
from jobflow.utils.find import (
    contains_flow_or_job,
    find_key,
    find_key_value,
    update_in_dictionary,
)
from jobflow.utils.log import initialize_logger
from jobflow.utils.uuid import suuid



"""Tools for finding and replacing in dictionaries and other objects."""

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing import Any, Hashable

    from monty.json import MSONable

__all__ = [
    "find_key",
    "find_key_value",
    "update_in_dictionary",
    "contains_flow_or_job",
]


def find_key(
    d: dict[Hashable, Any] | list[Any],
    key: Hashable | type[MSONable],
    include_end: bool = False,
    nested: bool = False,
) -> list[list[Any]]:
    """
    Find the route to key: value pairs in a dictionary.

    This function works on nested dictionaries, lists, and tuples.

    Parameters
    ----------
    d
        A dict or list of dicts.
    key
        A dictionary key or MSONable class to locate.
    include_end
        Whether to include the key in the route. This has no effect if the key is an
        MSONable class.
    nested:
        Whether to return nested keys or stop at the first match.

    Returns
    -------
    A list of routes to where the matches were found.

    Examples
    --------
    >>> data = {
    ...    "a": [0, {"b": 1, "x": 3}],
    ...    "c": {"d": {"x": 3}}
    ... }
    >>> find_key(data, "x")
    [['a', 1], ['c', 'd']]
    >>> find_key(data, "x", include_end=True)
    [['a', 1, 'x'], ['c', 'd', 'x']]

    The ``nested`` argument can be used to control the behaviour of nested keys.

    >>> data = {"a": {"x": {"x": 1}}, "b": {"x": 0}}
    >>> find_key(data, "x", nested=False)
    [['a'], ['b']]
    >>> find_key(data, "x", nested=True)
    [['a'], ['a', 'x'], ['b']]
    """
    import inspect

    from monty.json import MSONable

    found_items = set()

    def _lookup(obj, path=None):
        found = False
        if path is None:
            path = ()

        if isinstance(obj, dict):
            if (
                inspect.isclass(key)
                and issubclass(key, MSONable)
                and "@module" in obj
                and obj["@module"] == key.__module__
                and "@class" in obj
                and obj["@class"] == key.__name__
            ):
                found_items.add(path)
                found = True

            if key in obj:
                if include_end:
                    found_items.add((*path, key))
                else:
                    found_items.add(path)
                found = True

            if not found or nested:
                for k, v in obj.items():
                    _lookup(v, (*path, k))

        elif isinstance(obj, (list, tuple)):
            for i, v in enumerate(obj):
                _lookup(v, (*path, i))

    _lookup(d)
    return [list(path) for path in found_items]


def find_key_value(
    d: dict[Hashable, Any] | list[Any], key: Hashable, value: Hashable
) -> tuple[list[Any], ...]:
    """
    Find the route to key: value pairs in a dictionary.

    This function works on nested dictionaries, lists, and tuples.

    Parameters
    ----------
    d
        A dict or list of dicts.
    key
        A dictionary key.
    value
        A value.

    Returns
    -------
    A tuple of routes to where the matches were found.

    Examples
    --------
    >>> data = {
    ...    "a": [0, {"b": 1, "x": 3}],
    ...    "c": {"d": {"x": 3}}
    ... }
    ... find_key_value(data, "x", 3)
    (['a', 1], ['c', 'd'])
    """
    found_items = set()

    def _lookup(obj, path=None):
        if path is None:
            path = ()

        if isinstance(obj, dict):
            if key in obj and obj[key] == value:
                found_items.add(path)

            for k, v in obj.items():
                _lookup(v, (*path, k))

        elif isinstance(obj, (list, tuple)):
            for i, v in enumerate(obj):
                _lookup(v, (*path, i))

    _lookup(d)
    return tuple(list(path) for path in found_items)


def update_in_dictionary(obj: dict[Hashable, Any], updates: dict[tuple, Any]):
    """
    Update a dictionary in place at specific locations with a new values.

    This function works on nested dictionaries and those containing lists or tuples.

    Parameters
    ----------
    obj
        A dictionary to update.
    updates
        The updates to perform, as a dictionary of ``{location: update}``.

    Examples
    --------
    >>> data = {
    ...    "a": [0, {"b": 1, "x": 3}],
    ...    "c": {"d": {"x": 3}}
    ... }
    >>> update_in_dictionary(data, {('a', 1, 'x'): 100, ('c', 'd', 'x'): 100})
    >>> data
    {'a': [0, {'b': 1, 'x': 100}], 'c': {'d': {'x': 100}}}
    """
    for loc, update in updates.items():
        pos = obj
        for idx in loc[:-1]:
            pos = pos[idx]
        pos[loc[-1]] = update


def contains_flow_or_job(obj: Any) -> bool:
    """
    Find whether an object contains any :obj:`Flow` or :obj:`Job` objects.

    Parameters
    ----------
    obj
        An object.

    Returns
    -------
    bool
        Whether the object contains any Flows or jobs.
    """
    from monty.json import jsanitize

    from jobflow.core.flow import Flow
    from jobflow.core.job import Job

    if isinstance(obj, (Flow, Job)):
        # if the argument is an flow or job then stop there
        return True

    elif isinstance(obj, (float, int, str, bool)):
        # argument is a primitive, we won't find an flow or job here
        return False

    obj = jsanitize(obj, strict=True)

    # recursively find any reference classes
    locations = find_key_value(obj, "@class", "Flow")
    locations += find_key_value(obj, "@class", "Job")

    return len(locations) > 0



"""Define base Flow object."""

from __future__ import annotations

import logging
import typing
import warnings

from monty.json import MSONable

from jobflow.core.reference import find_and_get_references
from jobflow.utils import ValueEnum, contains_flow_or_job, suuid

if typing.TYPE_CHECKING:
    from typing import Any, Callable

    from networkx import DiGraph

    import jobflow

__all__ = ["JobOrder", "Flow", "get_flow"]

logger = logging.getLogger(__name__)


class JobOrder(ValueEnum):
    """
    Options to control the order of job execution.

    - ``AUTO``: Automatically determine the job order based on input and output
      references.
    - ``LINEAR``: Run the jobs in the order they appear in the jobs array.
    """

    AUTO = "auto"
    LINEAR = "linear"


class Flow(MSONable):
    """
    A Flow contains a collection of Jobs or other Flows to execute.

    The :obj:`Flow` object is the main tool for constructing workflows. Flows
    can either contain Jobs or other Flows. Like :obj:`.Job` objects, Flow objects
    can also have outputs, however, these are not explicitly stored in the database.
    Instead, the outputs of a Flow act to structure the outputs of the jobs
    contained within the Flow.

    Parameters
    ----------
    jobs
        The jobs to be run as a list of :obj:`.Job` or :obj:`Flow` objects.
    output
        The output of the flow. These should come from the output of one or more
        of the jobs.
    name
        The flow name.
    order
        The order in which the jobs should be executed. The default is to determine
        the order automatically based on the connections between jobs.
    uuid
        The identifier of the flow. This is generated automatically.
    hosts
        The list of UUIDs of the hosts containing the job. This is updated
        automatically when a flow is included in the jobs array of another flow.
        The object identified by one UUID of the list should be contained in objects
        identified by its subsequent elements.

    Raises
    ------
    ValueError
        If a job in the ``jobs`` array is already part of another flow.
    ValueError
        If any jobs needed to resolve the inputs of all jobs in the ``jobs`` array are
        missing.
    ValueError
        If any jobs needed to resolve the flow ``output`` are missing.

    Warns
    -----
    UserWarning
        If a ``.Job`` or ``Flow`` object is used as the Flow ``output`` rather
        than an ``OutputReference``.

    See Also
    --------
    .job, .Job, JobOrder

    Examples
    --------
    Below we define a simple job to add two numbers, and create an flow containing
    two connected add jobs.

    >>> from jobflow import job, Flow
    >>> @job
    ... def add(a, b):
    ...     return a + b
    >>> add_first = add(1, 2)
    >>> add_second = add(add_first.output, 2)
    >>> flow = Flow(jobs=[add_first, add_second])

    This flow does not expose any of the outputs of the jobs contained within it.
    We could instead "register" the output of the second add as the output of the
    flow.

    >>> flow = Flow(jobs=[add_first, add_second], output=add_second.output)

    This will allow the flow to be used in another flow. In this way, Flows
    can be infinitely nested. For example:

    >>> add_third = add(flow.output, 5)
    >>> outer_flow = Flow(jobs=[flow, add_third])

    Flows can be run using an flow manager. These enable running Flows
    locally or on compute clusters (using the FireWorks manager).

    >>> from jobflow.managers.local import run_locally
    >>> response = run_locally(flow)
    """

    def __init__(
        self,
        jobs: list[Flow | jobflow.Job] | jobflow.Job | Flow,
        output: Any | None = None,
        name: str = "Flow",
        order: JobOrder = JobOrder.AUTO,
        uuid: str = None,
        hosts: list[str] | None = None,
    ):
        from jobflow.core.job import Job

        if isinstance(jobs, (Job, Flow)):
            jobs = [jobs]

        if uuid is None:
            uuid = suuid()

        self.name = name
        self.order = order
        self.uuid = uuid
        self.hosts = hosts or []

        self._jobs: tuple[Flow | Job, ...] = ()
        self.add_jobs(jobs)
        self.output = output

    @property
    def jobs(self) -> tuple[Flow | jobflow.Job, ...]:
        """
        Get the Jobs in the Flow.

        Returns
        -------
        list[Job]
            The list of Jobs/Flows of the Flow.
        """
        return self._jobs

    @property
    def output(self) -> Any:
        """
        Get the output of the flow.

        Returns
        -------
        Any
            The output of the flow.
        """
        return self._output

    @output.setter
    def output(self, output: Any):
        """
        Set the output of the Flow.

        The output should be compatible with the list of Jobs/Flows contained in the
        Flow.

        Parameters
        ----------
        output
            The output of the flow. These should come from the output of one
            or more of the jobs.
        """
        if output is not None:
            if contains_flow_or_job(output):
                warnings.warn(
                    f"Flow '{self.name}' contains a Flow or Job as an output. "
                    f"Usually the Flow output should be the output of a Job or "
                    f"another Flow (e.g. job.output). If this message is "
                    f"unexpected then double check the outputs of your Flow."
                )

            # check if the jobs array contains all jobs needed for the references
            references = find_and_get_references(output)
            reference_uuids = {ref.uuid for ref in references}

            if not reference_uuids.issubset(set(self.job_uuids)):
                raise ValueError(
                    "jobs array does not contain all jobs needed for flow output"
                )
        self._output = output

    @property
    def job_uuids(self) -> tuple[str, ...]:
        """
        Uuids of every Job contained in the Flow (including nested Flows).

        Returns
        -------
        tuple[str]
            The uuids of all Jobs in the Flow (including nested Flows).
        """
        uuids: list[str] = []
        for job in self.jobs:
            if isinstance(job, Flow):
                uuids.extend(job.job_uuids)
            else:
                uuids.append(job.uuid)
        return tuple(uuids)

    @property
    def all_uuids(self) -> tuple[str, ...]:
        """
        Uuids of every Job and Flow contained in the Flow (including nested Flows).

        Returns
        -------
        tuple[str]
            The uuids of all Jobs and Flows in the Flow (including nested Flows).
        """
        uuids: list[str] = []
        for job in self.jobs:
            if isinstance(job, Flow):
                uuids.extend(job.all_uuids)
            uuids.append(job.uuid)
        return tuple(uuids)

    @property
    def graph(self) -> DiGraph:
        """
        Get a graph indicating the connectivity of jobs in the flow.

        Returns
        -------
        DiGraph
            The graph showing the connectivity of the jobs.
        """
        from itertools import product

        import networkx as nx

        graph = nx.compose_all([job.graph for job in self.jobs])

        if self.order == JobOrder.LINEAR:
            # add fake edges between jobs to force linear order
            edges = []
            for job_a, job_b in nx.utils.pairwise(self.jobs):
                if isinstance(job_a, Flow):
                    leaves = [v for v, d in job_a.graph.out_degree() if d == 0]
                else:
                    leaves = [job_a.uuid]

                if isinstance(job_b, Flow):
                    roots = [v for v, d in job_b.graph.in_degree() if d == 0]
                else:
                    roots = [job_b.uuid]

                for leaf, root in product(leaves, roots):
                    edges.append((leaf, root, {"properties": ""}))
            graph.add_edges_from(edges)
        return graph

    @property
    def host(self) -> str | None:
        """
        UUID of the first Flow that contains this Flow.

        Returns
        -------
        str
            the UUID of the host.
        """
        return self.hosts[0] if self.hosts else None

    def draw_graph(self, **kwargs):
        """
        Draw the flow graph using matplotlib.

        Requires matplotlib to be installed.

        Parameters
        ----------
        kwargs
            keyword arguments that are passed to :obj:`jobflow.utils.graph.draw_graph`.

        Returns
        -------
        pyplot
            The matplotlib pyplot state object.
        """
        from jobflow.utils.graph import draw_graph

        return draw_graph(self.graph, **kwargs)

    def iterflow(self):
        """
        Iterate through the jobs of the flow.

        The jobs are yielded such that the job output references can always be
        resolved. I.e., root nodes of the flow graph are always returned first.

        Yields
        ------
        Job, list[str]
            The Job and the uuids of any parent jobs (not to be confused with the host
            flow).
        """
        from networkx import is_directed_acyclic_graph

        from jobflow.utils.graph import itergraph

        graph = self.graph

        if not is_directed_acyclic_graph(graph):
            raise ValueError(
                "Job connectivity contains cycles therefore job execution order "
                "cannot be determined."
            )

        for node in itergraph(graph):
            parents = [u for u, v in graph.in_edges(node)]
            job = graph.nodes[node]["job"]
            yield job, parents

    def update_kwargs(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        function_filter: Callable | None = None,
        dict_mod: bool = False,
    ):
        """
        Update the kwargs of all Jobs in the Flow.

        Note that updates will be applied to jobs in nested Flow.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.

        Examples
        --------
        Consider a flow containing a simple job with a ``number`` keyword argument.

        >>> from jobflow import job, Flow
        >>> @job
        ... def add(a, number=5):
        ...     return a + number
        >>> add_job = add(1)
        >>> flow = Flow([add_job])

        The ``number`` argument could be updated in the following ways.

        >>> flow.update_kwargs({"number": 10})

        This will work if all jobs in the flow have a kwarg called number. However,
        when this is not the case this will result in the bad input kwargs for some
        jobs. To only apply the update to the correct jobs, filters can be used.

        >>> flow.update_kwargs({"number": 10}, name_filter="add")
        >>> flow.update_kwargs({"number": 10}, function_filter=add)
        """
        for job in self.jobs:
            job.update_kwargs(
                update,
                name_filter=name_filter,
                function_filter=function_filter,
                dict_mod=dict_mod,
            )

    def update_maker_kwargs(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        class_filter: type[jobflow.Maker] | None = None,
        nested: bool = True,
        dict_mod: bool = False,
    ):
        """
        Update the keyword arguments of any :obj:`.Maker` objects in the jobs.

        Note that updates will be applied to Jobs in any inner Flows.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the Maker name. Only Makers with a matching name will be
            updated. Includes partial matches, e.g. "ad" will match a Maker with the
            name "adder".
        class_filter
            A filter for the maker class. Only Makers with a matching class will be
            updated. Note the class filter will match any subclasses.
        nested
            Whether to apply the updates to Maker objects that are themselves kwargs
            of Maker, job, or flow objects. See examples for more details.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.

        Examples
        --------
        Consider the following flow containing jobs from a Maker:

        >>> from dataclasses import dataclass
        >>> from jobflow import job, Maker, Flow
        >>> @dataclass
        ... class AddMaker(Maker):
        ...     name: str = "add"
        ...     number: float = 10
        ...
        ...     @job
        ...     def make(self, a):
        ...         return a + self.number
        >>> maker = AddMaker()
        >>> add_job = maker.make(1)
        >>> flow = Flow([add_job])

        The ``number`` argument could be updated in the following ways.

        >>> flow.update_maker_kwargs({"number": 10})

        This will work if all Makers in the flow have a kwarg called number.
        However, when this is not the case this will result in the bad input kwargs
        for some Makers. To only apply the update to the correct Makers, filters can be
        used.

        >>> flow.update_maker_kwargs({"number": 10}, name_filter="add")
        >>> flow.update_maker_kwargs({"number": 10}, class_filter=AddMaker)

        By default, the updates are applied to nested Makers. These are Makers
        which are present in the kwargs of another Maker. Consider the following case
        for a Maker that produces a job that restarts.

        >>> from jobflow import Response
        >>> @dataclass
        ... class RestartMaker(Maker):
        ...     name: str = "replace"
        ...     add_maker: Maker = AddMaker()
        ...
        ...     @job
        ...     def make(self, a):
        ...         restart_job = self.add_maker.make(a)
        ...         return Response(replace=restart_job)
        >>> maker = RestartMaker()
        >>> my_job = maker.make(1)
        >>> flow = Flow([my_job]

        The following update will apply to the nested ``AddMaker`` in the kwargs of the
        ``RestartMaker``:

        >>> flow.update_maker_kwargs({"number": 10}, class_filter=AddMaker)

        However, if ``nested=False``, then the update will not be applied to the nested
        Maker:

        >>> flow.update_maker_kwargs(
        ...     {"number": 10}, class_filter=AddMaker, nested=False
        ... )
        """
        for job in self.jobs:
            job.update_maker_kwargs(
                update,
                name_filter=name_filter,
                class_filter=class_filter,
                nested=nested,
                dict_mod=dict_mod,
            )

    def append_name(self, append_str: str, prepend: bool = False):
        """
        Append a string to the name of the flow and all jobs contained in it.

        Parameters
        ----------
        append_str
            A string to append.
        prepend
            Prepend the name rather than appending it.
        """
        if prepend:
            self.name = append_str + self.name
        else:
            self.name += append_str

        for job in self.jobs:
            job.append_name(append_str, prepend=prepend)

    def update_metadata(
        self,
        update: dict[str, Any],
        name_filter: str | None = None,
        function_filter: Callable | None = None,
        dict_mod: bool = False,
        dynamic: bool = True,
    ):
        """
        Update the metadata of all Jobs in the Flow.

        Note that updates will be applied to jobs in nested Flow.

        Parameters
        ----------
        update
            The updates to apply.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        dict_mod
            Use the dict mod language to apply updates. See :obj:`.DictMods` for more
            details.
        dynamic
            The updates will be propagated to Jobs/Flows dynamically generated at
            runtime.

        Examples
        --------
        Consider a flow containing two jobs.

        >>> from jobflow import job, Flow
        >>> @job
        ... def add(a, b):
        ...     return a + b
        >>> add_job1 = add(5, 6)
        >>> add_job2 = add(6, 7)
        >>> flow = Flow([add_job1, add_job2])

        The ``metadata`` of both jobs could be updated as follows:

        >>> flow.update_metadata({"tag": "addition_job"})
        """
        for job in self.jobs:
            job.update_metadata(
                update,
                name_filter=name_filter,
                function_filter=function_filter,
                dict_mod=dict_mod,
                dynamic=dynamic,
            )

    def update_config(
        self,
        config: jobflow.JobConfig | dict,
        name_filter: str = None,
        function_filter: Callable = None,
        attributes: list[str] | str = None,
        dynamic: bool = True,
    ):
        """
        Update the job config of all Jobs in the Flow.

        Note that updates will be applied to jobs in nested Flow.

        Parameters
        ----------
        config
            A JobConfig object or a dict with containing the attributes to update.
        name_filter
            A filter for the job name. Only jobs with a matching name will be updated.
            Includes partial matches, e.g. "ad" will match a job with the name "adder".
        function_filter
            A filter for the job function. Only jobs with a matching function will be
            updated.
        attributes :
            Which attributes of the job config to set. Can be specified as one or more
            attributes specified by their name.
        dynamic
            The updates will be propagated to Jobs/Flows dynamically generated at
            runtime.

        Examples
        --------
        Consider a flow containing two jobs.

        >>> from jobflow import job, Flow
        >>> @job
        ... def add(a, b):
        ...     return a + b
        >>> add_job1 = add(5, 6)
        >>> add_job2 = add(6, 7)
        >>> flow = Flow([add_job1, add_job2])

        The ``config`` of both jobs could be updated as follows:

        >>> new_config = JobConfig(
        ...    manager_config={"_fworker": "myfworker"}, resolve_references=False
        ... )
        >>> flow.update_config(new_config)

        To only update specific attributes, the ``attributes`` argument can be set. For
        example, the following will only update the "manager_config" attribute of the
        jobs' config.

        >>> flow.update_config(new_config, attributes="manager_config")

        Alternatively, the config can be specified as a dictionary with keys that are
        attributes of the JobConfig object. This allows you to specify updates without
        having to create a completely new JobConfig object. For example:

        >>> flow.update_config({"manager_config": {"_fworker": "myfworker"}})
        """
        for job in self.jobs:
            job.update_config(
                config,
                name_filter=name_filter,
                function_filter=function_filter,
                attributes=attributes,
                dynamic=dynamic,
            )

    def add_hosts_uuids(
        self, hosts_uuids: str | list[str] | None = None, prepend: bool = False
    ):
        """
        Add a list of UUIDs to the internal list of hosts.

        If hosts_uuids is None the uuid of this Flow will be added to the inner jobs and
        flow. Otherwise, the passed value will be set both in the list of hosts
        of the current flow and of the inner jobs and flows.
        The elements of the list are supposed to be ordered in such a way that
        the object identified by one UUID of the list is contained in objects
        identified by its subsequent elements.

        Parameters
        ----------
        hosts_uuids
            A list of UUIDs to add. If None the current uuid of the flow will be
            added to the inner Flows and Jobs.
        prepend
            Insert the UUIDs at the beginning of the list rather than extending it.
        """
        if hosts_uuids is not None:
            if not isinstance(hosts_uuids, (list, tuple)):
                hosts_uuids = [hosts_uuids]
            if prepend:
                self.hosts[0:0] = hosts_uuids
            else:
                self.hosts.extend(hosts_uuids)
        else:
            hosts_uuids = [self.uuid]
        for j in self.jobs:
            j.add_hosts_uuids(hosts_uuids, prepend=prepend)

    def add_jobs(self, jobs: list[Flow | jobflow.Job] | jobflow.Job | Flow):
        """
        Add Jobs or Flows to the Flow.

        Added objects should not belong to other flows. The list of hosts will be added
        automatically to the incoming Jobs/Flows based on the hosts of the current Flow.

        Parameters
        ----------
        jobs
            A list of Jobs and Flows.
        """
        if not isinstance(jobs, (tuple, list)):
            jobs = [jobs]

        job_ids = set(self.all_uuids)
        hosts = [self.uuid, *self.hosts]
        for job in jobs:
            if job.host is not None and job.host != self.uuid:
                raise ValueError(
                    f"{job.__class__.__name__} {job.name} ({job.uuid}) already belongs "
                    f"to another flow."
                )
            if job.uuid in job_ids:
                raise ValueError(
                    "jobs array contains multiple jobs/flows with the same uuid "
                    f"({job.uuid})"
                )
            # check for circular dependency of Flows.
            if isinstance(job, Flow) and self.uuid in job.all_uuids:
                raise ValueError(
                    f"circular dependency: Flow ({job.uuid}) contains the "
                    f"current Flow ({self.uuid})"
                )
            job_ids.add(job.uuid)
            job.add_hosts_uuids(hosts)
        self._jobs += tuple(jobs)

    def remove_jobs(self, indices: int | list[int]):
        """
        Remove jobs from the Flow.

        It is not possible to remove jobs referenced in the output.

        Parameters
        ----------
        indices
            Indices of the jobs to be removed. Accepted values: from 0 to len(jobs) - 1.
        """
        if not isinstance(indices, (list, tuple)):
            indices = [indices]
        if any(i < 0 or i >= len(self.jobs) for i in indices):
            raise ValueError(
                "Only indices between 0 and the number of the jobs are accepted"
            )

        new_jobs = tuple(j for i, j in enumerate(self.jobs) if i not in indices)
        uuids: set = set()
        for job in new_jobs:
            if isinstance(job, Flow):
                uuids.update(job.job_uuids)
            else:
                uuids.add(job.uuid)

        # check if the output contains some references to the removed Jobs.
        references = find_and_get_references(self.output)
        reference_uuids = {ref.uuid for ref in references}

        if not reference_uuids.issubset(uuids):
            raise ValueError(
                "Removed Jobs/Flows are referenced in the output of the Flow."
            )

        self._jobs = new_jobs


def get_flow(
    flow: Flow | jobflow.Job | list[jobflow.Job],
) -> Flow:
    """
    Check dependencies and return flow object.

    Parameters
    ----------
    flow
        A job, list of jobs, or flow.

    Returns
    -------
    Flow
        A :obj:`Flow` object where connections have been checked.
    """
    if not isinstance(flow, Flow):
        flow = Flow(jobs=flow)

    # ensure that we have all the jobs needed to resolve the reference connections
    job_references = find_and_get_references(flow.jobs)
    job_reference_uuids = {ref.uuid for ref in job_references}
    missing_jobs = job_reference_uuids.difference(set(flow.job_uuids))
    if len(missing_jobs) > 0:
        raise ValueError(
            "The following jobs were not found in the jobs array and are needed to "
            f"resolve output references:\n{list(missing_jobs)}"
        )

    return flow



"""
Implementation of the DictMod language for manipulating dictionaries.

This module enables the modification of a dict using another dict. The main method of
interest is :obj:`apply_mod`.

.. Note::
    This code is based heavily on the Ansible class of `custodian
    <https://pypi.python.org/pypi/custodian>`_, but simplifies it considerably for the
    limited use cases required by jobflow.

    The original version of this file was written by Shyue Ping Ong and Anubhav Jain.
"""

from __future__ import annotations

import re
import typing

if typing.TYPE_CHECKING:
    from typing import Any


__all__ = ["DictMods", "apply_mod"]


class DictMods:
    """
    Class to define mongo-like modifications on a dict.

    Supported keywords include the following Mongo-based keywords, with the usual
    meanings (refer to Mongo documentation for information):

    - ``_inc``
    - ``_set``
    - ``_unset``
    - ``_push``
    - ``_push_all``
    - ``_add_to_set`` (but ``_each`` is not supported)
    - ``_pop``
    - ``_pull``
    - ``_pull_all``
    - ``_rename``

    .. Note::

        Note that ``_set`` does not support modification of nested dicts using the mongo
        ``{"a.b":1}`` notation. This is because mongo does not allow keys with "." to be
        inserted. Instead, nested dict modification is supported using a special "->"
        keyword, e.g. ``{"a->b": 1}``
    """

    def __init__(self):
        self.supported_actions = {}
        for i in dir(self):
            if (not re.match(r"__\w+__", i)) and callable(getattr(self, i)):
                self.supported_actions["_" + i] = getattr(self, i)

    @staticmethod
    def set(input_dict, settings):
        """Set a value."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            d[key] = v

    @staticmethod
    def unset(input_dict, settings):
        """Unset a value."""
        for k in settings:
            (d, key) = _get_nested_dict(input_dict, k)
            del d[key]

    @staticmethod
    def push(input_dict, settings):
        """Append a value to a list or create a list it if it doesn't exist."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d:
                d[key].append(v)
            else:
                d[key] = [v]

    @staticmethod
    def push_all(input_dict, settings):
        """Extend a list or create a list it if it doesn't exist."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d:
                d[key].extend(v)
            else:
                d[key] = v

    @staticmethod
    def inc(input_dict, settings):
        """Increment a number."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d:
                d[key] += v
            else:
                d[key] = v

    @staticmethod
    def rename(input_dict, settings):
        """Rename a key."""
        for k, v in settings.items():
            if k in input_dict:
                input_dict[v] = input_dict[k]
                del input_dict[k]

    @staticmethod
    def add_to_set(input_dict, settings):
        """Add an item to a set or create the set if it doesn't exist."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d and (not isinstance(d[key], (list, tuple))):
                raise ValueError(f"Keyword {k} does not refer to an array.")
            if key in d and v not in d[key]:
                d[key].append(v)
            elif key not in d:
                d[key] = v

    @staticmethod
    def pull(input_dict, settings):
        """Extract a field from a nested dictionary."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d and (not isinstance(d[key], (list, tuple))):
                raise ValueError(f"Keyword {k} does not refer to an array.")
            if key in d:
                d[key] = [i for i in d[key] if i != v]

    @staticmethod
    def pull_all(input_dict, settings):
        """Extract many fields from a nested dictionary."""
        for k, v in settings.items():
            if k in input_dict and (not isinstance(input_dict[k], (list, tuple))):
                raise ValueError(f"Keyword {k} does not refer to an array.")
            for i in v:
                DictMods.pull(input_dict, {k: i})

    @staticmethod
    def pop(input_dict, settings):
        """Pop a value from a list."""
        for k, v in settings.items():
            (d, key) = _get_nested_dict(input_dict, k)
            if key in d and (not isinstance(d[key], (list, tuple))):
                raise ValueError(f"Keyword {k} does not refer to an array.")
            if v == 1:
                d[key].pop()
            elif v == -1:
                d[key].pop(0)


_DM = DictMods()


def apply_mod(modification: dict[str, Any], obj: dict[str, Any]):
    """
    Apply a dict mod to an object.

    Note that modify makes actual in-place modifications. It does not return a copy.

    Parameters
    ----------
    modification
        Modification must be ``{action_keyword : settings}``, where action_keyword is a
        supported DictMod.
    obj
        A dict to be modified.
    """
    for action, settings in modification.items():
        if action in _DM.supported_actions:
            _DM.supported_actions[action].__call__(obj, settings)
        else:
            raise ValueError(f"{action} is not a supported action!")


def _get_nested_dict(
    input_dict: dict[str, Any], key: str
) -> tuple[dict[str, Any], str] | None:
    """Get nested dicts using a key."""
    current = input_dict
    toks = [t for t in key.split("->") if t != ""]
    n = len(toks)
    for i, tok in enumerate(toks):
        if tok not in current and i < n - 1:
            current[tok] = {}
        elif i == n - 1:
            return current, toks[-1]
        current = current[tok]
    return None
